#
# Command-line arguments handling tests for generator
#
# Copyright (C) 2018 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel.lapierre@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from .base import TestBase, exe_configure, exe_generate, OVS_CLEANUP


class TestConfigArgs(TestBase):
    '''Config file argument handling'''

    def test_invalid_args(self):
        # Systemd generator stage
        generator = os.path.join(self.workdir.name, 'usr', 'lib', 'systemd', 'system-generators', 'netplan')
        try:
            subprocess.check_output([generator, '--root-dir', self.workdir.name,
                                     self.generator_dir, self.generator_early_dir, self.generator_late_dir,
                                     '--invalid-arg'],
                                    stderr=subprocess.STDOUT, text=True)
        except subprocess.CalledProcessError as e:
            self.assertIn('failed to parse options: Unknown option --invalid-arg', e.output)
            self.assertEqual(e.returncode, 1)

        # Netplan configure stage
        try:
            subprocess.check_output([exe_configure, '--root-dir', self.workdir.name, '--invalid-arg'],
                                    stderr=subprocess.STDOUT, text=True)
        except subprocess.CalledProcessError as e:
            self.assertIn('failed to parse options: Unknown option --invalid-arg', e.output)
            self.assertEqual(e.returncode, 1)

    def test_no_files(self):
        # Systemd generator stage
        generator = os.path.join(self.workdir.name, 'usr', 'lib', 'systemd', 'system-generators', 'netplan')
        try:
            subprocess.check_output([generator, '--root-dir', self.workdir.name,
                                     self.generator_dir, self.generator_early_dir, self.generator_late_dir],
                                    stderr=subprocess.STDOUT, text=True)
        except subprocess.CalledProcessError as e:  # pragma: nocover (debug output)
            self.fail(f"systemd generator failed: {e.output.strip()}")

        # Netplan configure stage
        subprocess.check_call([exe_configure, '--root-dir', self.workdir.name])
        # nothing is generated in /run besides the netplan-ovs-cleanup.service.
        # We only have the sd-generator in usr/lib/systemd/system-generators
        self.assertEqual(set(os.listdir(self.workdir.name)), {'usr', 'run'})
        self.assert_networkd(None)
        self.assert_networkd_udev(None)
        self.assert_nm(None)
        self.assert_nm_udev(None)
        self.assert_ovs({'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})

    def test_no_configs(self):
        self.generate('network:\n  version: 2')
        # should not write any files
        self.assertEqual(set(os.listdir(self.workdir.name)) - {'usr'}, {'etc', 'run'})
        self.assert_networkd(None)
        self.assert_networkd_udev(None)
        self.assert_nm(None)
        self.assert_nm_udev(None)
        self.assert_ovs({'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # should not touch -wait-online
        service_dir = os.path.join(self.workdir.name, 'run', 'systemd', 'system')
        override = os.path.join(service_dir, 'systemd-networkd-wait-online.service.d', '10-netplan.conf')
        self.assertFalse(os.path.isfile(override))

    def test_empty_config(self):
        self.generate('')
        # should not write any files
        self.assertEqual(set(os.listdir(self.workdir.name)) - {'usr'}, {'etc', 'run'})
        self.assert_networkd(None)
        self.assert_networkd_udev(None)
        self.assert_nm(None)
        self.assert_nm_udev(None)
        self.assert_ovs({'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})

    def test_generate_fails_during_try(self):
        os.makedirs(self.rundir, mode=0o700, exist_ok=True)
        open(os.path.join(self.rundir, "netplan-try.ready"), "w").close()

        self.generate('network:\n  version: 2', expect_fail=True)

    def test_file_args(self):
        conf = os.path.join(self.workdir.name, 'config')
        with open(conf, 'w') as f:
            f.write('network: {}')
        # when specifying custom files, it should ignore the global config
        self.generate('''network:
  version: 2
  ethernets:
    eth0:
      dhcp4: true''', extra_args=[conf])
        # There is one systemd service unit 'netplan-ovs-cleanup.service' in /run,
        # which will always be created
        self.assertEqual(set(os.listdir(self.workdir.name)) - {'usr'}, {'config', 'etc', 'run'})
        self.assert_networkd(None)
        self.assert_networkd_udev(None)
        self.assert_nm(None)
        self.assert_nm_udev(None)

    def test_file_args_notfound(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth0:
      dhcp4: true''', expect_fail=True, extra_args=['/non/existing/config'])
        self.assertEqual(err, 'Cannot stat /non/existing/config: No such file or directory\n')
        # No backend configuration was generated
        self.assert_networkd(None)
        self.assert_networkd_udev(None)
        self.assert_nm(None)
        self.assert_nm_udev(None)

    def test_help(self):
        conf = os.path.join(self.workdir.name, 'etc', 'netplan', 'a.yaml')
        os.makedirs(os.path.dirname(conf))
        with open(conf, 'w') as f:
            f.write('''network:
  version: 2
  ethernets:
    eth0:
      dhcp4: true''')
        os.chmod(conf, mode=0o600)

        p = subprocess.Popen([exe_generate, '--root-dir', self.workdir.name, '--help'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             text=True)
        (out, err) = p.communicate()
        self.assertEqual(err, '')
        self.assertEqual(p.returncode, 0)
        self.assertIn('Usage:', out)
        self.assertEqual(set(os.listdir(self.workdir.name)) - {'usr'}, {'etc'})

    def test_unknown_cli_args(self):
        p = subprocess.Popen([exe_generate, '--foo'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             text=True)
        (out, err) = p.communicate()
        self.assertIn('nknown option --foo', err)
        self.assertNotEqual(p.returncode, 0)

    def test_output_mkdir_error(self):
        conf = os.path.join(self.workdir.name, 'config')
        with open(conf, 'w') as f:
            f.write('''network:
  version: 2
  ethernets:
    eth0:
      dhcp4: true''')
        err = self.generate('', extra_args=['--root-dir', '/proc/foo', conf], expect_fail=True)
        # can be /proc/foor/run/systemd/{network,system}
        self.assertIn('cannot create directory /proc/foo/run/systemd/', err)

    def test_systemd_generator(self):
        conf = os.path.join(self.confdir, 'a.yaml')
        os.makedirs(os.path.dirname(conf))
        with open(conf, 'w') as f:
            f.write('''network:
  version: 2
  ethernets:
    eth99:
      dhcp4: true
    eth98:
      dhcp4: true
      optional: true
    lo:
      addresses: ["127.0.0.1/8", "::1/128"]
  vlans:
    eth99.42:  # link-local is ignored for bridge-members
      link: eth99
      id: 42
      link-local: [ipv4, ipv6]
    eth99.43: # no link-local, but routable IP
      link: eth99
      id: 43
      link-local: []
      addresses: [10.0.0.2/24]
    eth99.44: # link-loal, but no routable IP
      link: eth99
      id: 44
      link-local: [ipv6]
    eth99.45: # ignore-carrier & no link-local, but routable IP
      link: eth99
      id: 45
      link-local: []
      ignore-carrier: true
      addresses: [10.0.0.3/24]
    eth99.46: # routable, but optional. So no wait-online log message about this bond member
      link: eth99
      id: 46
      dhcp4: true
      optional: true
  bonds:
    bond0:
      dhcp4: true
      interfaces: [eth99.42, eth99.43, eth99.46]''')
        os.chmod(conf, mode=0o600)

        generator = os.path.join(self.workdir.name, 'usr', 'lib', 'systemd', 'system-generators', 'netplan')
        local_env = os.environ.copy()
        local_env['G_MESSAGES_DEBUG'] = 'all'
        out = ''
        try:
            out = subprocess.check_output([generator, '--root-dir', self.workdir.name,
                                           self.generator_dir, self.generator_early_dir, self.generator_late_dir],
                                          stderr=subprocess.STDOUT, text=True, env=local_env)
        except subprocess.CalledProcessError as e:  # pragma: nocover (debug output)
            self.fail(f"systemd generator failed: {e.output.strip()}")
        self.assert_networkd(None)  # we're only executing the sd-generator here, not the ./configure stage

        # check log message about bonds wait-online
        self.assertIn('Not all bond members need to be connected for bond0 to be ready.', out)
        self.assertIn('Consider marking eth99.42 as "optional: true", to avoid blocking systemd-networkd-wait-online.', out)
        self.assertNotIn('making eth99.43 as "optional: true"', out)  # routable
        self.assertNotIn('making eth99.46 as "optional: true"', out)  # optional

        # should auto-enable networkd and -wait-online
        self.assertTrue(os.path.islink(os.path.join(
            self.generator_dir, 'multi-user.target.wants', 'systemd-networkd.service')))
        self.assertTrue(os.path.islink(os.path.join(
            self.generator_dir, 'network-online.target.wants', 'systemd-networkd-wait-online.service')))
        override = os.path.join(self.generator_late_dir, 'systemd-networkd-wait-online.service.d', '10-netplan.conf')
        self.assertTrue(os.path.isfile(override))
        with open(override, 'r') as f:
            # eth99 does not exist on the system, so will not be listed
            self.assertEqual(f.read(), '''[Unit]
ConditionPathIsSymbolicLink=/run/systemd/generator/network-online.target.wants/systemd-networkd-wait-online.service
After=systemd-resolved.service

[Service]
ExecStart=
ExecStart=/lib/systemd/systemd-networkd-wait-online -i eth99.43:carrier -i lo:carrier \
-i eth99.42:carrier -i eth99.44:degraded -i bond0:degraded
ExecStart=/lib/systemd/systemd-networkd-wait-online --any --dns -o routable -i eth99.43 -i eth99.45 -i bond0\n''')

    def test_systemd_generator_all_optional(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth0:
      dhcp4: true
      optional: true''')
        n = os.path.join(self.workdir.name, 'run', 'systemd', 'network', '10-netplan-eth0.network')
        self.assertTrue(os.path.exists(n))
        os.unlink(n)

        # should auto-enable networkd but not -wait-online
        self.assertTrue(os.path.islink(os.path.join(
            self.generator_dir, 'multi-user.target.wants', 'systemd-networkd.service')))
        self.assertFalse(os.path.islink(os.path.join(
            self.generator_dir, 'network-online.target.wants', 'systemd-networkd-wait-online.service')))
        override = os.path.join(self.generator_late_dir, 'systemd-networkd-wait-online.service.d', '10-netplan.conf')
        self.assertTrue(os.path.isfile(override))
        with open(override, 'r') as f:
            self.assertEqual(f.read(), '''[Unit]
ConditionPathIsSymbolicLink=/run/systemd/generator/network-online.target.wants/systemd-networkd-wait-online.service
''')

    def test_systemd_wait_online_only_non_routable(self):
        self.generate('''network:
  version: 2
  ethernets:
    nomatchfound: # non-optional, but cannot be matched to a physical interface on the test runner
      dhcp4: true
  vlans:
    eth99.44:
      link: nomatchfound
      id: 44
      link-local: [ipv6]''')
        override = os.path.join(self.generator_late_dir, 'systemd-networkd-wait-online.service.d', '10-netplan.conf')
        self.assertTrue(os.path.isfile(override))
        with open(override, 'r') as f:
            self.assertEqual(f.read(), '''[Unit]
ConditionPathIsSymbolicLink=/run/systemd/generator/network-online.target.wants/systemd-networkd-wait-online.service

[Service]
ExecStart=
ExecStart=/lib/systemd/systemd-networkd-wait-online -i eth99.44:degraded
''')

    def test_systemd_wait_online_only_routable(self):
        self.generate('''network:
  version: 2
  bridges:
    br0:
      dhcp4: true''')
        n = os.path.join(self.workdir.name, 'run', 'systemd', 'network', '10-netplan-br0.network')
        self.assertTrue(os.path.exists(n))
        os.unlink(n)

        override = os.path.join(self.generator_late_dir, 'systemd-networkd-wait-online.service.d', '10-netplan.conf')
        self.assertTrue(os.path.isfile(override))
        with open(override, 'r') as f:
            self.assertEqual(f.read(), '''[Unit]
ConditionPathIsSymbolicLink=/run/systemd/generator/network-online.target.wants/systemd-networkd-wait-online.service
After=systemd-resolved.service

[Service]
ExecStart=
ExecStart=/lib/systemd/systemd-networkd-wait-online -i br0:degraded
ExecStart=/lib/systemd/systemd-networkd-wait-online --any --dns -o routable -i br0
''')

    def test_systemd_generator_noconf(self):
        self.generate('')
        # no enablement symlink here (in multi-user.target.wants)
        self.assertFalse(os.path.exists(os.path.join(
            self.generator_late_dir, 'multi-user.target.wants', 'systemd-networkd.service')))

    def test_systemd_generator_badcall(self):
        outdir = os.path.join(self.workdir.name, 'out')
        os.mkdir(outdir)

        generator = os.path.join(self.workdir.name, 'usr', 'lib', 'systemd', 'system-generators', 'netplan')
        try:
            subprocess.check_output([generator, '--root-dir', self.workdir.name],
                                    stderr=subprocess.STDOUT, text=True)
            self.fail("direct systemd generator call is expected to fail, but succeeded.")  # pragma: nocover
        except subprocess.CalledProcessError as e:
            self.assertEqual(e.returncode, 1)
            self.assertIn('can not be called directly', e.output)

    def test_systemd_generator_escaping(self):
        conf = '''network:
  version: 2
  ethernets:
    lo:
      match:
        name: lo
      set-name: "a ; b\\t; c\\t; d \\n 123 ; echo "
      addresses: ["127.0.0.1/8", "::1/128"]'''
        self.generate(conf)

        n = os.path.join(self.workdir.name, 'run', 'systemd', 'network', '10-netplan-lo.network')
        self.assertTrue(os.path.exists(n))
        os.unlink(n)

        # should auto-enable networkd and -wait-online
        self.assertTrue(os.path.islink(os.path.join(
            self.generator_dir, 'multi-user.target.wants', 'systemd-networkd.service')))
        self.assertTrue(os.path.islink(os.path.join(
            self.generator_dir, 'network-online.target.wants', 'systemd-networkd-wait-online.service')))
        override = os.path.join(self.generator_late_dir, 'systemd-networkd-wait-online.service.d', '10-netplan.conf')
        self.assertTrue(os.path.isfile(override))
        with open(override, 'r') as f:
            # eth99 does not exist on the system, so will not be listed
            self.assertEqual(f.read(), '''[Unit]
ConditionPathIsSymbolicLink=/run/systemd/generator/network-online.target.wants/systemd-networkd-wait-online.service
After=systemd-resolved.service

[Service]
ExecStart=
ExecStart=/lib/systemd/systemd-networkd-wait-online -i a \\; b\\t; c\\t; d \\n 123 \\; echo :degraded
ExecStart=/lib/systemd/systemd-networkd-wait-online --any --dns -o routable -i a \\; b\\t; c\\t; d \\n 123 \\; echo \n''')



#!/usr/bin/python3
#
# Copyright (C) 2020-2023 Canonical, Ltd.
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''netplan get command line'''

from ..state import NetplanConfigState
from .. import utils


class NetplanGet(utils.NetplanCommand):
    def __init__(self):
        super().__init__(command_id='get',
                         description='Get a setting by specifying a nested key like "ethernets.eth0.addresses", or "all"',
                         leaf=True)

    def run(self):
        self.parser.add_argument('key', type=str, nargs='?', default='all', help='The nested key in dotted format')
        self.parser.add_argument('--root-dir', default='/',
                                 help='Read configuration files from this root directory instead of /')

        self.func = self.command_get

        self.parse_args()
        self.run_command()

    def command_get(self):
        state_data = NetplanConfigState(self.key, self.root_dir)
        print(state_data, end='')



#!/usr/bin/python3
#
# Regression tests to catch previously-fixed issues.
#
# These need to be run in a VM and do change the system
# configuration.
#
# Copyright (C) 2018-2021 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel-lapierre@canonical.com>
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import signal
import subprocess
import time
import unittest

from base import IntegrationTestsBase, test_backends


class _CommonTests():

    def test_empty_yaml_lp1795343(self):
        with open(self.config, 'w') as f:
            f.write('''''')
        self.generate_and_settle([])


@unittest.skipIf("networkd" not in test_backends,
                 "skipping as networkd backend tests are disabled")
class TestNetworkd(IntegrationTestsBase, _CommonTests):
    backend = 'networkd'

    def test_lp1802322_bond_mac_rename(self):
        self.setup_eth(None)
        self.addCleanup(subprocess.call, ['ip', 'link', 'delete', 'mybond'], stderr=subprocess.DEVNULL)
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  ethernets:
    ethbn1:
      match: {name: %(ec)s}
      dhcp4: no
    ethbn2:
      match: {name: %(e2c)s}
      dhcp4: no
  bonds:
    mybond:
      interfaces: [ethbn1, ethbn2]
      macaddress: 00:0a:f7:72:a7:28
      mtu: 9000
      addresses: [ 192.168.5.9/24 ]
      gateway4: 192.168.5.1
      parameters:
        down-delay: 0
        lacp-rate: fast
        mii-monitor-interval: 100
        mode: 802.3ad
        transmit-hash-policy: layer3+4
        up-delay: 0
      ''' % {'r': self.backend, 'ec': self.dev_e_client, 'e2c': self.dev_e2_client})
        self.generate_and_settle([self.dev_e_client, self.dev_e2_client, 'mybond'])
        self.assert_iface_up(self.dev_e_client,
                             ['master mybond', '00:0a:f7:72:a7:28'],  # wokeignore:rule=master
                             ['inet '])
        self.assert_iface_up(self.dev_e2_client,
                             ['master mybond', '00:0a:f7:72:a7:28'],  # wokeignore:rule=master
                             ['inet '])
        self.assert_iface_up('mybond', ['inet 192.168.5.[0-9]+/24'])
        with open('/sys/class/net/mybond/bonding/slaves') as f:  # wokeignore:rule=slave
            self.assertIn(self.dev_e_client, f.read().strip())

    def test_try_accept_lp1949095(self):
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  version: 2''' % {'r': self.backend})
        p = subprocess.Popen(['netplan', 'try'], bufsize=1, text=True,
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        time.sleep(5)
        p.send_signal(signal.SIGUSR1)
        out, err = p.communicate(timeout=10)
        self.assertEqual('', err)
        self.assertNotIn('An error occurred:', out)
        self.assertIn('Configuration accepted.', out)

    def test_try_reject_lp1949095(self):
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  version: 2''' % {'r': self.backend})
        p = subprocess.Popen(['netplan', 'try'], bufsize=1, text=True,
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        time.sleep(5)
        p.send_signal(signal.SIGINT)
        out, err = p.communicate(timeout=10)
        self.assertEqual('', err)
        self.assertNotIn('An error occurred:', out)
        self.assertIn('Reverting.', out)

    def test_apply_networkd_inactive_lp1962095(self):
        self.setup_eth(None)
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  ethernets:
    %(ec)s:
      dhcp4: true
    %(e2c)s:
      dhcp4: true
  version: 2''' % {'r': self.backend, 'ec': self.dev_e_client, 'e2c': self.dev_e2_client})
        # stop networkd to simulate the failure case
        subprocess.check_call(['systemctl', 'stop', 'systemd-networkd.service', 'systemd-networkd.socket'])
        self.generate_and_settle([self.state_dhcp4(self.dev_e_client), self.state_dhcp4(self.dev_e2_client)])
        self.assert_iface_up(self.dev_e_client, ['inet 192.168.5.[0-9]+/24'])
        self.assert_iface_up(self.dev_e2_client, ['inet 192.168.6.[0-9]+/24'])


@unittest.skipIf("NetworkManager" not in test_backends,
                 "skipping as NetworkManager backend tests are disabled")
class TestNetworkManager(IntegrationTestsBase, _CommonTests):
    backend = 'NetworkManager'

    def test_try_accept_lp1959570(self):
        original_env = dict(os.environ)
        self.addCleanup(subprocess.call, ['ip', 'link', 'delete', 'br54'], stderr=subprocess.DEVNULL)
        self.addCleanup(subprocess.call, ['mv', '/snap/bin/nmcli', '/usr/bin/nmcli'], stderr=subprocess.DEVNULL)
        self.addCleanup(os.environ.update, original_env)
        os.makedirs('/snap/bin', exist_ok=True)
        subprocess.call(['mv', '/usr/bin/nmcli', '/snap/bin/nmcli'])
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  version: 2
  bridges:
    br54:
      addresses:
      - "10.0.0.20/24"''' % {'r': self.backend})
        del os.environ['PATH']  # clear PATH, to test for LP: #1959570
        p = subprocess.Popen(['/usr/sbin/netplan', 'try'], bufsize=1, text=True,
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        time.sleep(5)
        p.send_signal(signal.SIGUSR1)
        out, err = p.communicate(timeout=10)
        os.environ = original_env
        self.assertEqual('', err)
        self.assertNotIn('An error occurred:', out)
        self.assertIn('Configuration accepted.', out)


unittest.main(testRunner=unittest.TextTestRunner(stream=sys.stdout, verbosity=2))



#
# Tests for IPv6 Router Advertisement (RA) override handling in
# netplan generator
#
# Copyright (C) 2024 Canonical, Ltd.
# Author: Khoo Hao Yit <khoohaoyit16@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import TestBase


class TestNetworkd(TestBase):

    def assert_ra_overrides_key_value(
        self,
        yaml_field_name,
        yaml_field_value,
        networkd_field_name,
        networkd_field_value
    ):
        yaml_config = '''\
network:
  version: 2
  ethernets:
    engreen:
      ra-overrides:
        %s: %s
''' % (yaml_field_name, yaml_field_value)
        networkd_config = '''\
[Match]
Name=engreen

[Network]
LinkLocalAddressing=ipv6
'''
        if networkd_field_value is not None:
            networkd_config += '''
[IPv6AcceptRA]
%s=%s
''' % (networkd_field_name, networkd_field_value)
        self.generate(yaml_config)
        self.assert_networkd({'engreen.network': networkd_config})

    def assert_ra_overrides_all_fields(self):
        yaml_config = '''\
network:
  version: 2
  ethernets:
    engreen:
      ra-overrides:
        use-dns: false
        use-domains: route
        table: 701
'''
        networkd_config = '''\
[Match]
Name=engreen

[Network]
LinkLocalAddressing=ipv6

[IPv6AcceptRA]
UseDNS=false
UseDomains=route
RouteTable=701
'''
        self.generate(yaml_config)
        self.assert_networkd({'engreen.network': networkd_config})

    def test_ra_overrides_use_dns(self):
        self.assert_ra_overrides_key_value('use-dns', 'false', 'UseDNS', 'false')
        self.assert_ra_overrides_key_value('use-dns', 'true', 'UseDNS', 'true')

    def test_ra_overrides_use_domains(self):
        self.assert_ra_overrides_key_value('use-domains', 'false', 'UseDomains', 'false')
        self.assert_ra_overrides_key_value('use-domains', 'true', 'UseDomains', 'true')
        self.assert_ra_overrides_key_value('use-domains', 'route', 'UseDomains', 'route')

    def test_ra_overrides_table(self):
        self.assert_ra_overrides_key_value('table', '727', 'RouteTable', '727')

    def test_ra_overrides_all_fields(self):
        self.assert_ra_overrides_all_fields()


class TestConfigErrors(TestBase):

    def test_ra_overrides_use_domains_invalid_options(self):
        err = self.generate('''\
network:
  version: 2
  ethernets:
    engreen:
      ra-overrides:
        use-domains: invalid-options
''', expect_fail=True)
        self.assertIn("Invalid use-domains options 'invalid-options', must be a boolean, or the special value 'route'.", err)



#!/usr/bin/python3
#
# Copyright (C) 2019 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel-lapierre@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''netplan info command line'''

from .. import utils
from ... import _features


class NetplanInfo(utils.NetplanCommand):

    def __init__(self):
        super().__init__(command_id='info',
                         description='Show available features',
                         leaf=True)

    def run(self):  # pragma: nocover (covered in autopkgtest)
        format_group = self.parser.add_mutually_exclusive_group(required=False)
        format_group.add_argument('--json', dest='version_format', action='store_const',
                                  const='json',
                                  help='Output version and features in JSON format')
        format_group.add_argument('--yaml', dest='version_format', action='store_const',
                                  const='yaml',
                                  help='Output version and features in YAML format')

        self.func = self.command_info
        self.parse_args()
        self.run_command()

    def command_info(self):

        netplan_version = {
            'netplan.io': {
                'website': 'https://netplan.io/',
            }
        }

        flags = _features.NETPLAN_FEATURE_FLAGS
        netplan_version['netplan.io'].update({'features': flags})

        # Default to output in YAML format.
        if self.version_format is None:
            self.version_format = 'yaml'

        if self.version_format == 'json':
            import json
            print(json.dumps(netplan_version, indent=2))

        elif self.version_format == 'yaml':
            print('''netplan.io:
  website: "{}"
  features:'''.format(netplan_version['netplan.io']['website']))
            for feature in _features.NETPLAN_FEATURE_FLAGS:
                print('  - ' + feature)



#!/usr/bin/python3
#
# Integration tests for wireless devices
#
# These need to be run in a VM and do change the system
# configuration.
#
# Copyright (C) 2018-2021 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel-lapierre@canonical.com>
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import subprocess
import unittest

from base import IntegrationTestsWifi, test_backends


class _CommonTests():

    @unittest.skip("Unsupported matching by driver / wifi matching makes this untestable for now")
    def test_mapping_for_driver(self):
        self.setup_ap('hw_mode=b\nchannel=1\nssid=fake net', None)
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  wifis:
    wifi_ifs:
      match:
        driver: mac80211_hwsim
      dhcp4: yes
      access-points:
        "fake net": {}
        decoy: {}''' % {'r': self.backend})
        self.generate_and_settle([self.state_dhcp4(self.dev_w_client)])
        p = subprocess.Popen(['netplan', 'generate', '--mapping', 'mac80211_hwsim'],
                             stdout=subprocess.PIPE)
        out = p.communicate()[0]
        self.assertEquals(p.returncode, 1)
        self.assertIn(b'mac80211_hwsim', out)

    def test_wifi_ipv4_open(self):
        self.setup_ap('hw_mode=b\nchannel=1\nssid=fake net', None)

        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  wifis:
    %(wc)s:
      dhcp4: yes
      access-points:
        "fake net": {}
        decoy: {}''' % {'r': self.backend, 'wc': self.dev_w_client})
        self.generate_and_settle([self.state_dhcp4(self.dev_w_client)])
        self.assert_iface_up(self.dev_w_client, ['inet 192.168.5.[0-9]+/24'])
        self.assertIn(b'default via 192.168.5.1',  # from DHCP
                      subprocess.check_output(['ip', 'route', 'show', 'dev', self.dev_w_client]))
        if self.backend == 'NetworkManager':
            out = subprocess.check_output(['nmcli', 'dev', 'show', self.dev_w_client],
                                          text=True)
            self.assertRegex(out, 'GENERAL.CONNECTION.*netplan-%s-fake net' % self.dev_w_client)
            self.assertRegex(out, 'IP4.DNS.*192.168.5.1')
        else:
            out = subprocess.check_output(['networkctl', 'status', self.dev_w_client],
                                          text=True)
            self.assertRegex(out, 'DNS.*192.168.5.1')

    def test_wifi_ipv4_wpa2(self):
        self.setup_ap('''hw_mode=g
channel=1
ssid=fake net
wpa=1
wpa_key_mgmt=WPA-PSK
wpa_pairwise=TKIP
wpa_passphrase=12345678
''', None)

        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  wifis:
    %(wc)s:
      dhcp4: yes
      access-points:
        "fake net":
          password: 12345678
        decoy: {}''' % {'r': self.backend, 'wc': self.dev_w_client})
        self.generate_and_settle([self.state_dhcp4(self.dev_w_client)])
        self.assert_iface_up(self.dev_w_client, ['inet 192.168.5.[0-9]+/24'])
        self.assertIn(b'default via 192.168.5.1',  # from DHCP
                      subprocess.check_output(['ip', 'route', 'show', 'dev', self.dev_w_client]))
        if self.backend == 'NetworkManager':
            out = subprocess.check_output(['nmcli', 'dev', 'show', self.dev_w_client],
                                          text=True)
            self.assertRegex(out, 'GENERAL.CONNECTION.*netplan-%s-fake net' % self.dev_w_client)
            self.assertRegex(out, 'IP4.DNS.*192.168.5.1')
        else:
            out = subprocess.check_output(['networkctl', 'status', self.dev_w_client],
                                          text=True)
            self.assertRegex(out, 'DNS.*192.168.5.1')

    def test_wifi_ipv4_wpa2_psk_sha256_only(self):
        self.setup_ap('''hw_mode=g
channel=1
ssid=fake net
wpa=2
wpa_key_mgmt=WPA-PSK-SHA256
wpa_pairwise=CCMP
ieee80211w=2
wpa_passphrase=12345678
''', None)

        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  wifis:
    %(wc)s:
      dhcp4: yes
      access-points:
        "fake net":
          auth:
            key-management: psk-sha256
            password: 12345678
        decoy: {}''' % {'r': self.backend, 'wc': self.dev_w_client})
        self.generate_and_settle([self.state_dhcp4(self.dev_w_client)])
        self.assert_iface_up(self.dev_w_client, ['inet 192.168.5.[0-9]+/24'])
        self.assertIn(b'default via 192.168.5.1',  # from DHCP
                      subprocess.check_output(['ip', 'route', 'show', 'dev', self.dev_w_client]))
        if self.backend == 'NetworkManager':
            out = subprocess.check_output(['nmcli', 'dev', 'show', self.dev_w_client],
                                          text=True)
            self.assertRegex(out, 'GENERAL.CONNECTION.*netplan-%s-fake net' % self.dev_w_client)
            self.assertRegex(out, 'IP4.DNS.*192.168.5.1')
        else:
            out = subprocess.check_output(['networkctl', 'status', self.dev_w_client],
                                          text=True)
            self.assertRegex(out, 'DNS.*192.168.5.1')

    def test_wifi_regdom(self):
        self.setup_ap('''hw_mode=g
channel=1
ssid=fake net
wpa=1
wpa_key_mgmt=WPA-PSK
wpa_pairwise=TKIP
wpa_passphrase=12345678
''', None)

        out = subprocess.check_output(['iw', 'reg', 'get'], text=True)
        self.assertNotIn('country GB', out)
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  wifis:
    %(wc)s:
      addresses: ["192.168.1.42/24"]
      regulatory-domain: GB
      access-points:
        "fake net":
          password: 12345678''' % {'r': self.backend, 'wc': self.dev_w_client})
        self.generate_and_settle([self.dev_w_client])
        self.assert_iface_up(self.dev_w_client, ['inet 192.168.1.42/24'])
        out = subprocess.check_output(['iw', 'reg', 'get'], text=True)
        self.assertIn('global\ncountry GB', out)


@unittest.skipIf("networkd" not in test_backends,
                 "skipping as networkd backend tests are disabled")
class TestNetworkd(IntegrationTestsWifi, _CommonTests):
    backend = 'networkd'


@unittest.skipIf("NetworkManager" not in test_backends,
                 "skipping as NetworkManager backend tests are disabled")
class TestNetworkManager(IntegrationTestsWifi, _CommonTests):
    backend = 'NetworkManager'

    def test_wifi_ap_open(self):
        # we use dev_w_client and dev_w_ap in switched roles here, to keep the
        # existing device denylisting in NM; i. e. dev_w_client is the
        # NM-managed AP, and dev_w_ap the manually managed client
        with open(self.config, 'w') as f:
            f.write('''network:
  wifis:
    renderer: NetworkManager
    %(wc)s:
      dhcp4: yes
      access-points:
        "fake net":
          mode: ap''' % {'wc': self.dev_w_client})
        self.generate_and_settle([self.state(self.dev_w_client, 'inet 10.')])
        out = subprocess.check_output(['iw', 'dev', self.dev_w_client, 'info'],
                                      text=True)
        self.assertIn('type AP', out)
        self.assertIn('ssid fake net', out)

        # connect the other end
        subprocess.check_call(['ip', 'link', 'set', self.dev_w_ap, 'up'])
        subprocess.check_call(['iw', 'dev', self.dev_w_ap, 'connect', 'fake net'])
        out = subprocess.check_output(['dhcpcd', '-w', '-d', self.dev_w_ap],
                                      stderr=subprocess.STDOUT, text=True)
        self.assertIn('acknowledged 10.', out)
        out = subprocess.check_output(['iw', 'dev', self.dev_w_ap, 'info'],
                                      text=True)
        self.assertIn('type managed', out)
        self.assertIn('ssid fake net', out)
        self.assert_iface_up(self.dev_w_ap, ['inet 10.'])

    @unittest.skip("Test if flaky. NM might generate a different MAC address.")
    def test_wifi_cloned_macaddress_stable_ssid(self):
        self.setup_ap('''hw_mode=g
channel=1
ssid=fake net
wpa=1
wpa_key_mgmt=WPA-PSK
wpa_pairwise=TKIP
wpa_passphrase=12345678
''', None)

        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: NetworkManager
  wifis:
    %(wc)s:
      addresses: ["192.168.1.42/24"]
      dhcp4: false
      dhcp6: false
      macaddress: stable-ssid
      access-points:
        "fake net":
          password: 12345678''' % {'wc': self.dev_w_client})

        subprocess.check_call(['systemctl', 'start', 'NetworkManager'])

        # Make the generated MAC address predictable
        # See nm_utils_hw_addr_gen_stable_eth() in NM for details
        # TODO: save and restore these files to avoid any impact on the
        # entire test suite.
        with open('/etc/machine-id', 'w') as f:
            f.write('ee7ac3602b6306061bd984a41eb1c045\n')
        with open('/var/lib/NetworkManager/secret_key', 'w') as f:
            f.write('nm-v2:hnIHoHp4p9kaEWU5/+dO+gFREirN1AsMoO1MPaoYxCc=')

        subprocess.check_call(['systemctl', 'restart', 'NetworkManager'])

        self.generate_and_settle([self.state_up(self.dev_w_client)])
        self.assert_iface_up(self.dev_w_client, ['ether 5e:ba:fe:fd:89:03'])


unittest.main(testRunner=unittest.TextTestRunner(stream=sys.stdout, verbosity=2))



#
# Tests for network authentication config generated via netplan
#
# Copyright (C) 2018 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel.lapierre@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import stat

from .base import TestBase, ND_DHCP4, ND_WIFI_DHCP4, SD_WPA, NM_MANAGED, NM_UNMANAGED


class TestNetworkd(TestBase):

    def test_auth_wifi_detailed(self):
        self.generate('''network:
  version: 2
  wifis:
    wl0:
      access-points:
        "Joe's Home":
          password: "s0s3kr1t"
        "Luke's Home":
          auth:
            key-management: psk
            password: "4lsos3kr1t"
        "BobsHome":
          password: "e03ce667c87bc81ca968d9120ca37f89eb09aec3c55b80386e5d772efd6b926e"
        "BillsHome":
          auth:
            key-management: psk
            password: "db3b0acf5653aeaddd5fe034fb9f07175b2864f847b005aaa2f09182d9411b04"
        workplace:
          auth:
            key-management: eap
            method: ttls
            anonymous-identity: "@internal.example.com"
            identity: "joe@internal.example.com"
            password: "v3ryS3kr1t"
        workplace2:
          auth:
            key-management: eap
            method: peap
            identity: "joe@internal.example.com"
            password: "v3ryS3kr1t"
            ca-certificate: /etc/ssl/work2-cacrt.pem
        workplacehashed:
          auth:
            key-management: eap
            method: ttls
            anonymous-identity: "@internal.example.com"
            identity: "joe@internal.example.com"
            password: hash:9db1636cedc5948537e7bee0cc1e9590
        customernet:
          auth:
            key-management: eap
            method: tls
            anonymous-identity: "@cust.example.com"
            identity: "cert-joe@cust.example.com"
            ca-certificate: /etc/ssl/cust-cacrt.pem
            client-certificate: /etc/ssl/cust-crt.pem
            client-key: /etc/ssl/cust-key.pem
            client-key-password: "d3cryptPr1v4t3K3y"
        opennet:
          auth:
            key-management: none
        peer2peer:
          mode: adhoc
          auth: {}
      dhcp4: yes
      ''')

        self.assert_networkd({'wl0.network': ND_WIFI_DHCP4 % 'wl0'})
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'wl0')

        # generates wpa config and enables wpasupplicant unit
        with open(os.path.join(self.workdir.name, 'run/netplan/wpa-wl0.conf')) as f:
            new_config = f.read()
            self.assertIn('ctrl_interface=/run/wpa_supplicant', new_config)
            self.assertIn('''
network={
  ssid=P"peer2peer"
  mode=1
  key_mgmt=NONE
}
''', new_config)
            self.assertIn('''
network={
  ssid=P"Luke's Home"
  key_mgmt=WPA-PSK
  psk="4lsos3kr1t"
}
''', new_config)
            self.assertIn('''
network={
  ssid=P"BobsHome"
  key_mgmt=WPA-PSK WPA-PSK-SHA256 SAE
  ieee80211w=1
  psk=e03ce667c87bc81ca968d9120ca37f89eb09aec3c55b80386e5d772efd6b926e
}
''', new_config)
            self.assertIn('''
network={
  ssid=P"BillsHome"
  key_mgmt=WPA-PSK
  psk=db3b0acf5653aeaddd5fe034fb9f07175b2864f847b005aaa2f09182d9411b04
}
''', new_config)
            self.assertIn('''
network={
  ssid=P"workplace2"
  key_mgmt=WPA-EAP
  eap=PEAP
  identity="joe@internal.example.com"
  password="v3ryS3kr1t"
  ca_cert="/etc/ssl/work2-cacrt.pem"
}
''', new_config)
            self.assertIn('''
network={
  ssid=P"workplace"
  key_mgmt=WPA-EAP
  eap=TTLS
  identity="joe@internal.example.com"
  anonymous_identity="@internal.example.com"
  password="v3ryS3kr1t"
}
''', new_config)
            self.assertIn('''
network={
  ssid=P"workplacehashed"
  key_mgmt=WPA-EAP
  eap=TTLS
  identity="joe@internal.example.com"
  anonymous_identity="@internal.example.com"
  password=hash:9db1636cedc5948537e7bee0cc1e9590
}
''', new_config)
            self.assertIn('''
network={
  ssid=P"customernet"
  key_mgmt=WPA-EAP
  eap=TLS
  identity="cert-joe@cust.example.com"
  anonymous_identity="@cust.example.com"
  ca_cert="/etc/ssl/cust-cacrt.pem"
  client_cert="/etc/ssl/cust-crt.pem"
  private_key="/etc/ssl/cust-key.pem"
  private_key_passwd="d3cryptPr1v4t3K3y"
}
''', new_config)
            self.assertIn('''
network={
  ssid=P"opennet"
  key_mgmt=NONE
}
''', new_config)
            self.assertIn('''
network={
  ssid=P"Joe's Home"
  key_mgmt=WPA-PSK WPA-PSK-SHA256 SAE
  ieee80211w=1
  psk="s0s3kr1t"
}
''', new_config)
            self.assertEqual(stat.S_IMODE(os.fstat(f.fileno()).st_mode), 0o600)
        self.assertTrue(os.path.isfile(os.path.join(
            self.generator_late_dir, 'netplan-wpa-wl0.service')))
        self.assertTrue(os.path.islink(os.path.join(
            self.generator_late_dir, 'systemd-networkd.service.wants/netplan-wpa-wl0.service')))

    def test_auth_wired(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth0:
      auth:
        key-management: 802.1x
        method: tls
        anonymous-identity: "@cust.example.com"
        identity: "cert-joe@cust.example.com"
        ca-certificate: /etc/ssl/cust-cacrt.pem
        client-certificate: /etc/ssl/cust-crt.pem
        client-key: /etc/ssl/cust-key.pem
        client-key-password: "d3cryptPr1v4t3K3y"
        phase2-auth: MSCHAPV2
      dhcp4: yes
      ''')

        self.assert_networkd({'eth0.network': ND_DHCP4 % 'eth0'})
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'eth0')

        # generates wpa config and enables wpasupplicant unit
        with open(os.path.join(self.workdir.name, 'run/netplan/wpa-eth0.conf')) as f:
            self.assertEqual(f.read(), '''ctrl_interface=/run/wpa_supplicant

network={
  key_mgmt=IEEE8021X
  eap=TLS
  identity="cert-joe@cust.example.com"
  anonymous_identity="@cust.example.com"
  ca_cert="/etc/ssl/cust-cacrt.pem"
  client_cert="/etc/ssl/cust-crt.pem"
  private_key="/etc/ssl/cust-key.pem"
  private_key_passwd="d3cryptPr1v4t3K3y"
  phase2="auth=MSCHAPV2"
}
''')
            self.assertEqual(stat.S_IMODE(os.fstat(f.fileno()).st_mode), 0o600)
        self.assertTrue(os.path.isfile(os.path.join(
            self.generator_late_dir, 'netplan-wpa-eth0.service')))

        with open(os.path.join(self.generator_late_dir, 'netplan-wpa-eth0.service')) as f:
            self.assertEqual(f.read(), SD_WPA % {'iface': 'eth0', 'drivers': 'wired'})
            self.assertEqual(stat.S_IMODE(os.fstat(f.fileno()).st_mode), 0o644)
        self.assertTrue(os.path.islink(os.path.join(
            self.generator_late_dir, 'systemd-networkd.service.wants/netplan-wpa-eth0.service')))


class TestNetworkManager(TestBase):

    def test_auth_wifi_detailed(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  wifis:
    wl0:
      access-points:
        "Joe's Home":
          password: "s0s3kr1t"
        "Luke's Home":
          auth:
            key-management: psk
            password: "4lsos3kr1t"
        workplace:
          auth:
            key-management: eap
            method: ttls
            anonymous-identity: "@internal.example.com"
            identity: "joe@internal.example.com"
            password: "v3ryS3kr1t"
        workplace2:
          auth:
            key-management: eap
            method: peap
            identity: "joe@internal.example.com"
            password: "v3ryS3kr1t"
            ca-certificate: /etc/ssl/work2-cacrt.pem
        workplacehashed:
          auth:
            key-management: eap
            method: ttls
            anonymous-identity: "@internal.example.com"
            identity: "joe@internal.example.com"
            password: hash:9db1636cedc5948537e7bee0cc1e9590
        customernet:
          auth:
            key-management: 802.1x
            method: tls
            anonymous-identity: "@cust.example.com"
            identity: "cert-joe@cust.example.com"
            ca-certificate: /etc/ssl/cust-cacrt.pem
            client-certificate: /etc/ssl/cust-crt.pem
            client-key: /etc/ssl/cust-key.pem
            client-key-password: "d3cryptPr1v4t3K3y"
            phase2-auth: MSCHAPV2
        opennet:
          auth:
            key-management: none
        peer2peer:
          mode: adhoc
          auth: {}
      dhcp4: yes
      ''')

        self.assert_networkd({})
        self.assert_nm({'wl0-Joe%27s%20Home': '''[connection]
id=netplan-wl0-Joe's Home
type=wifi
interface-name=wl0

[ipv4]
method=auto

[ipv6]
method=ignore

[wifi]
ssid=Joe's Home
mode=infrastructure

[wifi-security]
key-mgmt=wpa-psk
pmf=2
psk=s0s3kr1t
''',
                        'wl0-Luke%27s%20Home': '''[connection]
id=netplan-wl0-Luke's Home
type=wifi
interface-name=wl0

[ipv4]
method=auto

[ipv6]
method=ignore

[wifi]
ssid=Luke's Home
mode=infrastructure

[wifi-security]
key-mgmt=wpa-psk
psk=4lsos3kr1t
''',
                        'wl0-workplace': '''[connection]
id=netplan-wl0-workplace
type=wifi
interface-name=wl0

[ipv4]
method=auto

[ipv6]
method=ignore

[wifi]
ssid=workplace
mode=infrastructure

[wifi-security]
key-mgmt=wpa-eap

[802-1x]
eap=ttls
identity=joe@internal.example.com
anonymous-identity=@internal.example.com
password=v3ryS3kr1t
''',
                        'wl0-workplace2': '''[connection]
id=netplan-wl0-workplace2
type=wifi
interface-name=wl0

[ipv4]
method=auto

[ipv6]
method=ignore

[wifi]
ssid=workplace2
mode=infrastructure

[wifi-security]
key-mgmt=wpa-eap

[802-1x]
eap=peap
identity=joe@internal.example.com
password=v3ryS3kr1t
ca-cert=/etc/ssl/work2-cacrt.pem
''',
                        'wl0-workplacehashed': '''[connection]
id=netplan-wl0-workplacehashed
type=wifi
interface-name=wl0

[ipv4]
method=auto

[ipv6]
method=ignore

[wifi]
ssid=workplacehashed
mode=infrastructure

[wifi-security]
key-mgmt=wpa-eap

[802-1x]
eap=ttls
identity=joe@internal.example.com
anonymous-identity=@internal.example.com
password=hash:9db1636cedc5948537e7bee0cc1e9590
''',
                        'wl0-customernet': '''[connection]
id=netplan-wl0-customernet
type=wifi
interface-name=wl0

[ipv4]
method=auto

[ipv6]
method=ignore

[wifi]
ssid=customernet
mode=infrastructure

[wifi-security]
key-mgmt=ieee8021x

[802-1x]
eap=tls
identity=cert-joe@cust.example.com
anonymous-identity=@cust.example.com
ca-cert=/etc/ssl/cust-cacrt.pem
client-cert=/etc/ssl/cust-crt.pem
private-key=/etc/ssl/cust-key.pem
private-key-password=d3cryptPr1v4t3K3y
phase2-auth=MSCHAPV2
''',
                        'wl0-opennet': '''[connection]
id=netplan-wl0-opennet
type=wifi
interface-name=wl0

[ipv4]
method=auto

[ipv6]
method=ignore

[wifi]
ssid=opennet
mode=infrastructure
''',
                        'wl0-peer2peer': '''[connection]
id=netplan-wl0-peer2peer
type=wifi
interface-name=wl0

[ipv4]
method=auto

[ipv6]
method=ignore

[wifi]
ssid=peer2peer
mode=adhoc
'''})
        self.assert_nm_udev(NM_MANAGED % 'wl0')

    def test_auth_wired(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    eth0:
      auth:
        key-management: 802.1x
        method: tls
        anonymous-identity: "@cust.example.com"
        identity: "cert-joe@cust.example.com"
        ca-certificate: /etc/ssl/cust-cacrt.pem
        client-certificate: /etc/ssl/cust-crt.pem
        client-key: /etc/ssl/cust-key.pem
        client-key-password: "d3cryptPr1v4t3K3y"
      dhcp4: yes
      ''')

        self.assert_nm({'eth0': '''[connection]
id=netplan-eth0
type=ethernet
interface-name=eth0

[ethernet]
wake-on-lan=0

[ipv4]
method=auto

[ipv6]
method=ignore

[802-1x]
eap=tls
identity=cert-joe@cust.example.com
anonymous-identity=@cust.example.com
ca-cert=/etc/ssl/cust-cacrt.pem
client-cert=/etc/ssl/cust-crt.pem
private-key=/etc/ssl/cust-key.pem
private-key-password=d3cryptPr1v4t3K3y
'''})
        self.assert_networkd({})
        self.assert_nm_udev(NM_MANAGED % 'eth0')


class TestConfigErrors(TestBase):

    def test_auth_invalid_key_mgmt(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth0:
      auth:
        key-management: bogus''', expect_fail=True)
        self.assertIn("unknown key management type 'bogus'", err)

    def test_auth_invalid_eap_method(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth0:
      auth:
        method: bogus''', expect_fail=True)
        self.assertIn("unknown EAP method 'bogus'", err)

    def test_auth_networkd_wifi_psk_too_big(self):
        err = self.generate('''network:
  version: 2
  wifis:
    wl0:
      access-points:
        "Joe's Home":
          password: "LoremipsumdolorsitametconsecteturadipiscingelitCrastemporvelitnunc"
      dhcp4: yes''', expect_fail=True)
        self.assertIn("ASCII passphrase must be between 8 and 63 characters (inclusive)", err)

    def test_auth_networkd_wifi_psk_too_small(self):
        err = self.generate('''network:
  version: 2
  wifis:
    wl0:
      access-points:
        "Joe's Home":
          password: "p4ss"
      dhcp4: yes''', expect_fail=True)
        self.assertIn("ASCII passphrase must be between 8 and 63 characters (inclusive)", err)

    def test_auth_networkd_wifi_psk_64_non_hexdigit(self):
        err = self.generate('''network:
  version: 2
  wifis:
    wl0:
      access-points:
        "Joe's Home":
          password: "LoremipsumdolorsitametconsecteturadipiscingelitCrastemporvelitnu"
      dhcp4: yes''', expect_fail=True)
        self.assertIn("PSK length of 64 is only supported for hex-digit representation", err)

    def test_auth_networkd_wire_psk_64_non_hexdigit(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth0:
      auth:
        key-management: psk
        password: "LoremipsumdolorsitametconsecteturadipiscingelitCrastemporvelitnu"
      dhcp4: yes''', expect_fail=True)
        self.assertIn("PSK length of 64 is only supported for hex-digit representation", err)



#
# Tests for bridge devices config generated via netplan
#
# Copyright (C) 2018 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel.lapierre@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from .base import TestBase, NM_UNMANAGED, NM_MANAGED


class TestNetworkd(TestBase):

    def test_bridge_set_mac(self):
        self.generate('''network:
  version: 2
  bridges:
    br0:
      macaddress: 00:01:02:03:04:05
      dhcp4: true''')

        self.assert_networkd({'br0.network': '''[Match]
Name=br0

[Link]
MACAddress=00:01:02:03:04:05

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6
ConfigureWithoutCarrier=yes

[DHCP]
RouteMetric=100
UseMTU=true
''',
                              'br0.netdev': '[NetDev]\nName=br0\nMACAddress=00:01:02:03:04:05\nKind=bridge\n'})

    def test_bridge_dhcp6_no_accept_ra(self):
        self.generate('''network:
  version: 2
  ethernets:
    engreen:
      dhcp4: no
      dhcp6: no
      accept-ra: no
  bridges:
    br0:
      interfaces: [engreen]
      dhcp6: true
      accept-ra: no''')
        self.assert_networkd({'br0.network': '''[Match]
Name=br0

[Network]
DHCP=ipv6
LinkLocalAddressing=ipv6
IPv6AcceptRA=no
ConfigureWithoutCarrier=yes

[DHCP]
RouteMetric=100
UseMTU=true
''',
                              'br0.netdev': '''[NetDev]
Name=br0
Kind=bridge
''',
                              'engreen.network': '''[Match]
Name=engreen

[Network]
LinkLocalAddressing=no
IPv6AcceptRA=no
Bridge=br0
'''})

    def test_bridge_empty(self):
        self.generate('''network:
  version: 2
  bridges:
    br0:
      dhcp4: true''')

        self.assert_networkd({'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n',
                              'br0.network': '''[Match]
Name=br0

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6
ConfigureWithoutCarrier=yes

[DHCP]
RouteMetric=100
UseMTU=true
'''})
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'br0')

    def test_bridge_type_renderer(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  bridges:
    renderer: networkd
    br0:
      dhcp4: true''')

        self.assert_networkd({'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n',
                              'br0.network': '''[Match]
Name=br0

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6
ConfigureWithoutCarrier=yes

[DHCP]
RouteMetric=100
UseMTU=true
'''})
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'br0')

    def test_bridge_def_renderer(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  bridges:
    renderer: NetworkManager
    br0:
      renderer: networkd
      addresses: [1.2.3.4/12]
      dhcp4: true''')

        self.assert_networkd({'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n',
                              'br0.network': '''[Match]
Name=br0

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6
Address=1.2.3.4/12
ConfigureWithoutCarrier=yes

[DHCP]
RouteMetric=100
UseMTU=true
'''})
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'br0')

    def test_bridge_forward_declaration(self):
        self.generate('''network:
  version: 2
  bridges:
    br0:
      interfaces: [eno1, switchports]
      dhcp4: true
  ethernets:
    eno1: {}
    switchports:
      match:
        driver: yayroute
''')

        self.assert_networkd({'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n',
                              'br0.network': '''[Match]
Name=br0

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6
ConfigureWithoutCarrier=yes

[DHCP]
RouteMetric=100
UseMTU=true
''',
                              'eno1.network': '[Match]\nName=eno1\n\n'
                                              '[Network]\nLinkLocalAddressing=no\nBridge=br0\n',
                              'switchports.network': '[Match]\nDriver=yayroute\n\n'
                                                     '[Network]\nLinkLocalAddressing=no\nBridge=br0\n'})

    @unittest.skipIf("CODECOV_TOKEN" in os.environ, "Skip on codecov.io: GLib changed hashtable sorting")
    def test_eth_bridge_nm_denylist(self):  # pragma: nocover
        self.generate('''network:
  renderer: networkd
  ethernets:
    eth42:
      dhcp4: yes
    ethbr:
      match: {name: eth43}
  bridges:
    mybr:
      interfaces: [ethbr]
      dhcp4: yes''')
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'eth42' + NM_UNMANAGED % 'eth43' + NM_UNMANAGED % 'mybr')

    def test_bridge_components(self):
        self.generate('''network:
  version: 2
  ethernets:
    eno1: {}
    switchports:
      match:
        driver: yayroute
  bridges:
    br0:
      interfaces: [eno1, switchports]
      dhcp4: true''')

        self.assert_networkd({'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n',
                              'br0.network': '''[Match]
Name=br0

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6
ConfigureWithoutCarrier=yes

[DHCP]
RouteMetric=100
UseMTU=true
''',
                              'eno1.network': '[Match]\nName=eno1\n\n'
                                              '[Network]\nLinkLocalAddressing=no\nBridge=br0\n',
                              'switchports.network': '[Match]\nDriver=yayroute\n\n'
                                                     '[Network]\nLinkLocalAddressing=no\nBridge=br0\n'})

    def test_bridge_params(self):
        self.generate('''network:
  version: 2
  ethernets:
    eno1: {}
    switchports:
      match:
        driver: yayroute
  bridges:
    br0:
      interfaces: [eno1, switchports]
      parameters:
        ageing-time: 50
        forward-delay: 12
        hello-time: 6
        max-age: 24
        priority: 1000
        stp: true
        path-cost:
          eno1: 70
        port-priority:
          eno1: 14
      dhcp4: true''')

        self.assert_networkd({'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n\n'
                                            '[Bridge]\nAgeingTimeSec=50\n'
                                            'Priority=1000\n'
                                            'ForwardDelaySec=12\n'
                                            'HelloTimeSec=6\n'
                                            'MaxAgeSec=24\n'
                                            'STP=true\n',
                              'br0.network': '''[Match]
Name=br0

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6
ConfigureWithoutCarrier=yes

[DHCP]
RouteMetric=100
UseMTU=true
''',
                              'eno1.network': '[Match]\nName=eno1\n\n'
                                              '[Network]\nLinkLocalAddressing=no\nBridge=br0\n\n'
                                              '[Bridge]\nCost=70\nPriority=14\n',
                              'switchports.network': '[Match]\nDriver=yayroute\n\n'
                                                     '[Network]\nLinkLocalAddressing=no\nBridge=br0\n'})


class TestNetworkManager(TestBase):

    def test_bridge_empty(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  bridges:
    br0:
      dhcp4: true''')

        self.assert_nm({'br0': '''[connection]
id=netplan-br0
type=bridge
interface-name=br0

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})
        self.assert_nm_udev(NM_MANAGED % 'br0')

    def test_bridge_type_renderer(self):
        self.generate('''network:
  version: 2
  renderer: networkd
  bridges:
    renderer: NetworkManager
    br0:
      dhcp4: true''')

        self.assert_nm({'br0': '''[connection]
id=netplan-br0
type=bridge
interface-name=br0

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})
        self.assert_nm_udev(NM_MANAGED % 'br0')

    def test_bridge_set_mac(self):
        self.generate('''network:
  version: 2
  bridges:
    renderer: NetworkManager
    br0:
      macaddress: 00:01:02:03:04:05
      dhcp4: true''')

        self.assert_nm({'br0': '''[connection]
id=netplan-br0
type=bridge
interface-name=br0

[ethernet]
cloned-mac-address=00:01:02:03:04:05

[ipv4]
method=auto

[ipv6]
method=ignore
'''})

    def test_bridge_def_renderer(self):
        self.generate('''network:
  version: 2
  renderer: networkd
  bridges:
    renderer: networkd
    br0:
      renderer: NetworkManager
      addresses: [1.2.3.4/12]
      dhcp4: true''')

        self.assert_nm({'br0': '''[connection]
id=netplan-br0
type=bridge
interface-name=br0

[ipv4]
method=auto
address1=1.2.3.4/12

[ipv6]
method=ignore
'''})
        self.assert_networkd({})
        self.assert_nm_udev(NM_MANAGED % 'br0')

    def test_bridge_forward_declaration(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  bridges:
    br0:
      interfaces: [eno1, switchport]
      dhcp4: true
  ethernets:
    eno1: {}
    switchport:
      match:
        name: enp2s1
''')

        self.assert_nm({'eno1': '''[connection]
id=netplan-eno1
type=ethernet
interface-name=eno1
slave-type=bridge # wokeignore:rule=slave
master=br0 # wokeignore:rule=master

[ethernet]
wake-on-lan=0

[ipv4]
method=link-local

[ipv6]
method=ignore
''',
                        'switchport': '''[connection]
id=netplan-switchport
type=ethernet
interface-name=enp2s1
slave-type=bridge # wokeignore:rule=slave
master=br0 # wokeignore:rule=master

[ethernet]
wake-on-lan=0

[ipv4]
method=link-local

[ipv6]
method=ignore
''',
                        'br0': '''[connection]
id=netplan-br0
type=bridge
interface-name=br0

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})
        self.assert_nm_udev(NM_MANAGED % 'br0' + NM_MANAGED % 'eno1' + NM_MANAGED % 'enp2s1')

    def test_bridge_components(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    eno1: {}
    switchport:
      match:
        name: enp2s1
  bridges:
    br0:
      interfaces: [eno1, switchport]
      dhcp4: true''')

        self.assert_nm({'eno1': '''[connection]
id=netplan-eno1
type=ethernet
interface-name=eno1
slave-type=bridge # wokeignore:rule=slave
master=br0 # wokeignore:rule=master

[ethernet]
wake-on-lan=0

[ipv4]
method=link-local

[ipv6]
method=ignore
''',
                        'switchport': '''[connection]
id=netplan-switchport
type=ethernet
interface-name=enp2s1
slave-type=bridge # wokeignore:rule=slave
master=br0 # wokeignore:rule=master

[ethernet]
wake-on-lan=0

[ipv4]
method=link-local

[ipv6]
method=ignore
''',
                        'br0': '''[connection]
id=netplan-br0
type=bridge
interface-name=br0

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})
        self.assert_nm_udev(NM_MANAGED % 'eno1' + NM_MANAGED % 'enp2s1' + NM_MANAGED % 'br0')

    def test_bridge_params(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    eno1: {}
    switchport:
      match:
        name: enp2s1
  bridges:
    br0:
      interfaces: [eno1, switchport]
      parameters:
        ageing-time: 50
        priority: 1000
        forward-delay: 12
        hello-time: 6
        max-age: 24
        path-cost:
          eno1: 70
        port-priority:
          eno1: 61
        stp: true
      dhcp4: true''')

        self.assert_nm({'eno1': '''[connection]
id=netplan-eno1
type=ethernet
interface-name=eno1
slave-type=bridge # wokeignore:rule=slave
master=br0 # wokeignore:rule=master

[bridge-port]
path-cost=70
priority=61

[ethernet]
wake-on-lan=0

[ipv4]
method=link-local

[ipv6]
method=ignore
''',
                        'switchport': '''[connection]
id=netplan-switchport
type=ethernet
interface-name=enp2s1
slave-type=bridge # wokeignore:rule=slave
master=br0 # wokeignore:rule=master

[ethernet]
wake-on-lan=0

[ipv4]
method=link-local

[ipv6]
method=ignore
''',
                        'br0': '''[connection]
id=netplan-br0
type=bridge
interface-name=br0

[bridge]
ageing-time=50
priority=1000
forward-delay=12
hello-time=6
max-age=24
stp=true

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})
        self.assert_nm_udev(NM_MANAGED % 'eno1' + NM_MANAGED % 'enp2s1' + NM_MANAGED % 'br0')


class TestNetplanYAMLv2(TestBase):
    '''No asserts are needed.

    The generate() method implicitly checks the (re-)generated YAML.
    '''

    def test_bridge_stp(self):
        self.generate('''network:
  version: 2
  bridges:
    br0:
      parameters:
        stp: no
      dhcp4: true''')


class TestConfigErrors(TestBase):

    def test_bridge_unknown_iface(self):
        err = self.generate('''network:
  version: 2
  bridges:
    br0:
      interfaces: ['foo']''', expect_fail=True)
        self.assertIn("br0: interface 'foo' is not defined", err)

    def test_bridge_multiple_assignments(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eno1: {}
  bridges:
    br0:
      interfaces: [eno1]
    br1:
      interfaces: [eno1]''', expect_fail=True)
        self.assertIn("br1: interface 'eno1' is already assigned to bridge br0", err)

    def test_bridge_invalid_dev_for_path_cost(self):
        self.generate('''network:
  version: 2
  ethernets:
    eno1:
      match:
        name: eth0
  bridges:
    br0:
      interfaces: [eno1]
      parameters:
        path-cost:
          eth0: 50
      dhcp4: true''', expect_fail=True)

    def test_bridge_path_cost_already_defined(self):
        self.generate('''network:
  version: 2
  ethernets:
    eno1:
      match:
        name: eth0
  bridges:
    br0:
      interfaces: [eno1]
      parameters:
        path-cost:
          eno1: 50
          eno1: 40
      dhcp4: true''', expect_fail=True)

    def test_bridge_invalid_path_cost(self):
        self.generate('''network:
  version: 2
  ethernets:
    eno1:
      match:
        name: eth0
  bridges:
    br0:
      interfaces: [eno1]
      parameters:
        path-cost:
          eno1: aa
      dhcp4: true''', expect_fail=True)

    def test_bridge_invalid_dev_for_port_prio(self):
        self.generate('''network:
  version: 2
  ethernets:
    eno1:
      match:
        name: eth0
  bridges:
    br0:
      interfaces: [eno1]
      parameters:
        port-priority:
          eth0: 50
      dhcp4: true''', expect_fail=True)

    def test_bridge_port_prio_already_defined(self):
        self.generate('''network:
  version: 2
  ethernets:
    eno1:
      match:
        name: eth0
  bridges:
    br0:
      interfaces: [eno1]
      parameters:
        port-priority:
          eno1: 50
          eno1: 40
      dhcp4: true''', expect_fail=True)

    def test_bridge_invalid_port_prio(self):
        self.generate('''network:
  version: 2
  ethernets:
    eno1:
      match:
        name: eth0
  bridges:
    br0:
      interfaces: [eno1]
      parameters:
        port-priority:
          eno1: 257
      dhcp4: true''', expect_fail=True)



#!/usr/bin/python3
#
# Copyright (C) 2023 Canonical, Ltd.
# Authors: Danilo Egea Gondolfo <danilo.egea.gondolfo@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
import ipaddress
import json
from socket import AF_INET, AF_INET6
from typing import AbstractSet

from netplan.netdef import NetplanRoute
from netplan_cli.cli.state import SystemConfigState, NetplanConfigState, DEVICE_TYPES
from netplan_cli.cli.utils import is_valid_macaddress, route_table_lookup


class DiffJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, NetplanRoute):
            return obj.to_dict()

        # Shouldn't be reached as the only non-serializable type we have at the moment is NetplanRoute
        return json.JSONEncoder.default(self, obj)  # pragma: nocover (only NetplanRoute requires the encoder)


class NetplanDiffState():
    '''
    DiffState is mainly responsible for getting both system's and Netplan's configuration
    state, compare them and provide a data-structure containing the differences it found.
    '''

    def __init__(self, system_state: SystemConfigState, netplan_state: NetplanConfigState):
        self.system_state = system_state
        self.netplan_state = netplan_state

        self.route_lookup_table_names = {}

    def get_full_state(self) -> dict:
        '''
        Return the states of both the system and Netplan in a common representation
        that makes it easier to compare them.
        '''

        full_state = {
            'interfaces': {}
        }

        system_interfaces = self._get_system_interfaces()
        netplan_interfaces = self._get_netplan_interfaces()

        # Merge all the interfaces in the same data structure
        all_interfaces = set(list(system_interfaces.keys()) + list(netplan_interfaces.keys()))
        for interface in all_interfaces:
            full_state['interfaces'][interface] = {}

        for interface, config in system_interfaces.items():
            full_state['interfaces'][interface].update(config)

        for interface, config in netplan_interfaces.items():
            full_state['interfaces'][interface].update(config)

        for interface in self.system_state.interface_list:
            full_state['interfaces'][interface.name]['data_sources'] = interface.data_sources

        return full_state

    def get_diff(self, interface: str = '') -> dict:
        '''
        Compare the configuration of interfaces currently found in the system against Netplan configuration.
        A number of heuristics are used to eliminate configuration that is automatically set in the system,
        such as certain routes and IP addresses. That is necessary because this configuration will not be found
        in Netplan. For example, if Netplan is enabling DHCP on an interface and not defining any extra IP addresses,
        we don't count the IPs automatically assigned to the interface as a difference. We do though count the eventual
        absence of addresses that should be assigned by DHCP as a difference.
        '''

        full_state = self.get_full_state()
        interfaces = self._get_comparable_interfaces(full_state.get('interfaces', {}))

        if interface:
            if config := interfaces.get(interface):
                interfaces = {interface: config}
            else:
                interfaces = {}

        report = self._create_new_report()

        self._analyze_missing_interfaces(report, interface)

        for interface, config in interfaces.items():
            netdef_id = config.get('system_state', {}).get('id')
            index = config.get('system_state', {}).get('index')
            iface = self._create_new_iface(netdef_id, interface, index)

            self._analyze_ip_addresses(config, iface)
            self._analyze_nameservers(config, iface)
            self._analyze_search_domains(config, iface)
            self._analyze_mac_addresses(config, iface)
            self._analyze_routes(config, iface)
            self._analyze_parent_links(config, iface)

            report['interfaces'].update(iface)

        # Sort the list of interfaces according to their indices.
        report['interfaces'] = dict(sorted(report['interfaces'].items(), key=lambda iface: iface[1].get('index')))
        return report

    def _create_new_report(self) -> dict:
        return {
            'interfaces': {},
            'missing_interfaces_system': {},
            'missing_interfaces_netplan': {},
        }

    def _create_new_iface(self, netdef_id: str, interface: str, index: int) -> dict:
        return {
            interface: {
                'index': index,
                'name': interface,
                'id': netdef_id,
                'system_state': {},
                'netplan_state': {},
            }
        }

    def _analyze_ip_addresses(self, config: dict, iface: dict) -> None:
        name = list(iface.keys())[0]
        netplan_ips = {ip for ip in config.get('netplan_state', {}).get('addresses', [])}
        netplan_ips = self._normalize_ip_addresses(netplan_ips)

        missing_dhcp4_address = config.get('netplan_state', {}).get('dhcp4', False)
        missing_dhcp6_address = config.get('netplan_state', {}).get('dhcp6', False)
        link_local = config.get('netplan_state', {}).get('link_local', [])
        system_ips = set()
        for addr, addr_data in config.get('system_state', {}).get('addresses', {}).items():
            ip = ipaddress.ip_interface(addr)
            flags = addr_data.get('flags', [])

            # Select only static IPs
            if not {'dhcp', 'dynamic', 'link', 'ra'}.intersection(flags):
                system_ips.add(addr)

            # Handle the link local address
            # If it's present but the respective setting is not enabled in the netdef
            # it's considered a difference.
            if 'link' in flags and ip.is_link_local:
                if isinstance(ip.ip, ipaddress.IPv4Address) and 'ipv4' not in link_local:
                    system_ips.add(addr)
                if isinstance(ip.ip, ipaddress.IPv6Address) and 'ipv6' not in link_local:
                    system_ips.add(addr)

            # We assume the IPv4 was acquired via DHCP if any of the flags
            # 'dhcp' or 'dynamic' is set
            # For IPv6, DHCP is not the only way to acquire a dynamic address,
            # so we only check for the 'dhcp' flag
            if isinstance(ip.ip, ipaddress.IPv4Address):
                if {'dhcp', 'dynamic'}.intersection(flags):
                    missing_dhcp4_address = False
            if isinstance(ip.ip, ipaddress.IPv6Address):
                if 'dhcp' in flags:
                    missing_dhcp6_address = False

            # Handle RA addresses
            # We only consider the presence of addresses with the 'ra' flag to be a diff
            # if 'accept_ra' is set to false.
            # When it's unset, networkd will enable it by default.
            accept_ra = config.get('netplan_state', {}).get('accept_ra')
            if 'ra' in flags and accept_ra is False:
                system_ips.add(addr)

        present_only_in_netplan = netplan_ips.difference(system_ips)
        present_only_in_system = system_ips.difference(netplan_ips)

        if missing_dhcp4_address:
            iface[name]['system_state']['missing_dhcp4_address'] = True

        if missing_dhcp6_address:
            iface[name]['system_state']['missing_dhcp6_address'] = True

        if present_only_in_system:
            iface[name]['netplan_state'].update({
                'missing_addresses': list(sorted(present_only_in_system)),
            })

        if present_only_in_netplan:
            iface[name]['system_state'].update({
                'missing_addresses': list(sorted(present_only_in_netplan)),
            })

    def _get_comparable_interfaces(self, interfaces: dict) -> dict:
        ''' In order to compare interfaces, they must exist in the system AND in Netplan.
            Here we filter out interfaces that don't have a system_state, a netplan_state
            or a netdef ID.

            There is a special case where the interface will have a system_state and a netdef_id
            but will be missing in Netplan. That will happen when the user removes the interface
            only from Netplan but doesn't run netplan apply.
        '''
        filtered = {}

        for interface, config in interfaces.items():
            if config.get('system_state') is None or config.get('netplan_state') is None:
                continue

            if not config.get('system_state', {}).get('id'):
                continue

            filtered[interface] = config

        return filtered

    def _normalize_ip_addresses(self, addresses: set) -> set:
        ''' Apply some transformations to IP addresses so their representation
        will match the system's.
        '''
        new_ips_set = set()
        for ip in addresses:
            ip = self._compress_ipv6_address(ip)
            new_ips_set.add(ip)

        return new_ips_set

    def _compress_ipv6_address(self, address: str) -> str:
        '''
        Compress IPv6 addresses to match the system representation
        Example: 1:2:0:0::123/64 -> 1:2::123/64
                 1:2:0:0::123 -> 1:2::123
        If "address" is not an IPv6Address, return the original value
        '''
        try:
            addr = ipaddress.ip_interface(address)
            if '/' in address:
                return addr.with_prefixlen
            return str(addr.ip)
        except ValueError:
            return address

    def _analyze_nameservers(self, config: dict, iface: dict) -> None:
        name = list(iface.keys())[0]
        data_sources = config.get('data_sources', {}).get('dns', {})

        # TODO: improve analysis of configuration received from DHCP
        # when Network Manager is used

        netplan_nameservers = set(config.get('netplan_state', {}).get('nameservers_addresses', []))
        system_nameservers = set(config.get('system_state', {}).get('nameservers_addresses', []))

        dhcp4_ns = {ns for ns in data_sources if data_sources[ns] == 'DHCPv4'}
        dhcp6_ns = {ns for ns in data_sources if data_sources[ns] == 'DHCPv6'}
        ndisc_ns = {ns for ns in data_sources if data_sources[ns] == 'NDisc'}

        # Filter out dynamically assigned DNS data
        accept_ra_enabled = config.get('netplan_state', {}).get('accept_ra') is not False
        link_local_enabled = 'ipv6' in config.get('netplan_state', {}).get('link_local', [])

        if dhcp4_ns or dhcp6_ns or ndisc_ns:
            if config.get('netplan_state', {}).get('dhcp4'):
                system_nameservers = system_nameservers - dhcp4_ns

            # networkd will enable DHCPv6 if RA is enabled even if DHCPv6 is set to false
            # see systemd.network(5)
            if accept_ra_enabled:
                system_nameservers = system_nameservers - dhcp6_ns

            link_local_ns = {ns for ns in system_nameservers if ns in ndisc_ns}
            if link_local_enabled and accept_ra_enabled:
                system_nameservers = system_nameservers - link_local_ns
        else:
            # Heuristic used when we don't have data sources for DNS information (when
            # Network Manager is used for example)
            # If the netplan configuration has DHCP enabled and an empty list of nameservers
            # we assume it's dynamic. So we don't take these addresses into account when comparing
            # them, except for the ones we already identified as RA/LL.
            # Note: Some useful information can be found in /var/run/systemd/netif/leases/
            # but the lease files have a comment saying they shouldn't be parsed.
            # There is a feature request to expose more DHCP information via the DBus API
            # https://github.com/systemd/systemd/issues/27699
            if not netplan_nameservers:
                if config.get('netplan_state', {}).get('dhcp4'):
                    system_nameservers = {ns for ns in system_nameservers
                                          if not isinstance(ipaddress.ip_address(ns), ipaddress.IPv4Address)}
                if config.get('netplan_state', {}).get('dhcp6'):
                    system_nameservers = {ns for ns in system_nameservers
                                          if not isinstance(ipaddress.ip_address(ns), ipaddress.IPv6Address)}

        present_only_in_netplan = netplan_nameservers.difference(system_nameservers)
        present_only_in_system = system_nameservers.difference(netplan_nameservers)

        if present_only_in_system:
            iface[name]['netplan_state'].update({
                'missing_nameservers_addresses': list(present_only_in_system),
            })

        if present_only_in_netplan:
            iface[name]['system_state'].update({
                'missing_nameservers_addresses': list(present_only_in_netplan),
            })

    def _analyze_search_domains(self, config: dict, iface: dict) -> None:
        name = list(iface.keys())[0]
        netplan_search_domains = set(config.get('netplan_state', {}).get('nameservers_search', []))
        system_search_domains = set(config.get('system_state', {}).get('nameservers_search', []))
        data_sources = config.get('data_sources', {}).get('search', {})
        dhcp4_search = {s for s in data_sources if data_sources[s] == 'DHCPv4'}
        dhcp6_search = {s for s in data_sources if data_sources[s] == 'DHCPv6'}

        if dhcp4_search or dhcp6_search:
            if config.get('netplan_state', {}).get('dhcp4'):
                system_search_domains = system_search_domains - dhcp4_search

            # networkd will enable DHCPv6 if RA is enabled even if DHCPv6 is set to false
            # see systemd.network(5)
            accept_ra_enabled = config.get('netplan_state', {}).get('accept_ra') is not False
            if accept_ra_enabled:
                system_search_domains = system_search_domains - dhcp6_search
        else:
            # Heuristic used when we don't have data sources for DNS information (when
            # Network Manager is used for example)
            # If the netplan configuration has DHCP enabled and an empty list of search domains
            # we assume it's dynamic
            if not netplan_search_domains:
                if config.get('netplan_state', {}).get('dhcp4') or config.get('netplan_state', {}).get('dhcp6'):
                    system_search_domains = set()

        present_only_in_netplan = netplan_search_domains.difference(system_search_domains)
        present_only_in_system = system_search_domains.difference(netplan_search_domains)

        if present_only_in_system:
            iface[name]['netplan_state'].update({
                'missing_nameservers_search': list(present_only_in_system),
            })

        if present_only_in_netplan:
            iface[name]['system_state'].update({
                'missing_nameservers_search': list(present_only_in_netplan),
            })

    def _analyze_mac_addresses(self, config: dict, iface: dict) -> None:
        name = list(iface.keys())[0]
        system_macaddress = config.get('system_state', {}).get('macaddress')
        netplan_macaddress = config.get('netplan_state', {}).get('macaddress')

        # if the macaddress in netplan is an special option (such as 'random')
        # don't try to diff it against the system MAC address
        if netplan_macaddress and not is_valid_macaddress(netplan_macaddress):
            return

        if system_macaddress and netplan_macaddress:
            if system_macaddress != netplan_macaddress:
                iface[name]['system_state'].update({
                    'missing_macaddress': netplan_macaddress
                })
                iface[name]['netplan_state'].update({
                    'missing_macaddress': system_macaddress
                })

    def _analyze_routes(self, config: dict, iface: dict) -> None:
        name = list(iface.keys())[0]
        netplan_routes = set(config.get('netplan_state', {}).get('routes', []))
        system_routes = set(config.get('system_state', {}).get('routes', []))
        netplan_routes = self._normalize_routes(netplan_routes)

        # If gateway4 and/or gateway6 is present, merge them in the routes set
        if netplan_gateways := self._normalize_gateways(config.get('netplan_state', {})):
            netplan_routes.update(netplan_gateways)

        # Filter out some routes that are expected to be added automatically
        system_addresses = [ip for ip in config.get('system_state', {}).get('addresses', {})]
        system_routes = self._filter_system_routes(system_routes, system_addresses, config)

        present_only_in_netplan = netplan_routes.difference(system_routes)
        present_only_in_system = system_routes.difference(netplan_routes)

        if present_only_in_system:
            iface[name]['netplan_state'].update({
                'missing_routes': [route for route in sorted(present_only_in_system, key=lambda r: r.to)],
            })

        if present_only_in_netplan:
            iface[name]['system_state'].update({
                'missing_routes': [route for route in sorted(present_only_in_netplan, key=lambda r: r.to)],
            })

    def _analyze_missing_interfaces(self, report: dict, interface: str) -> None:
        netplan_interfaces = {iface for iface in self.netplan_state.netdefs}
        system_interfaces_netdef_ids = {iface.netdef_id for iface in self.system_state.interface_list if iface.netdef_id}

        netplan_only = netplan_interfaces.difference(system_interfaces_netdef_ids)
        # Filtering out disconnected wifi netdefs
        # If a wifi netdef is present in the netplan_only list it's because it's disconnected
        netplan_only = list(filter(lambda i: self.netplan_state.netdefs.get(i).type != 'wifis', netplan_only))

        system_only = []
        for iface in self.system_state.interface_list:
            if iface.netdef_id not in netplan_interfaces:
                system_only.append(iface.name)

        netplan_only = sorted(netplan_only)
        system_only = sorted(system_only)

        if interface:
            netplan_only = filter(lambda i: i == interface, netplan_only)
            system_only = filter(lambda i: i == interface, system_only)

        system_state = self.system_state.get_data()

        for iface in netplan_only:
            iface_type = self.netplan_state.netdefs.get(iface).type
            report['missing_interfaces_system'][iface] = {
                'type': DEVICE_TYPES.get(iface_type, 'other')
            }

        for iface in system_only:
            report['missing_interfaces_netplan'][iface] = {
                'type': system_state.get(iface).get('type', 'other'),
                'index': system_state.get(iface).get('index'),
            }

    def _analyze_parent_links(self, config: dict, iface: dict) -> None:
        '''
        Analyse if interfaces such as bonds, bridges and VRFs are correctly attached to their
        members and vice versa.
        '''
        name = list(iface.keys())[0]
        bond = [config.get('system_state', {}).get('bond'), config.get('netplan_state', {}).get('bond')]
        bridge = [config.get('system_state', {}).get('bridge'), config.get('netplan_state', {}).get('bridge')]
        vrf = [config.get('system_state', {}).get('vrf'), config.get('netplan_state', {}).get('vrf')]
        interfaces = [config.get('system_state', {}).get('interfaces', []), config.get('netplan_state', {}).get('interfaces', [])]

        if bond != [None, None] and bond[0] != bond[1]:
            if bond[0]:
                iface[name]['netplan_state']['missing_bond_link'] = bond[0]
            if bond[1]:
                iface[name]['system_state']['missing_bond_link'] = bond[1]

        if bridge != [None, None] and bridge[0] != bridge[1]:
            if bridge[0]:
                iface[name]['netplan_state']['missing_bridge_link'] = bridge[0]
            if bridge[1]:
                iface[name]['system_state']['missing_bridge_link'] = bridge[1]

        if vrf != [None, None] and vrf[0] != vrf[1]:
            if vrf[0]:
                iface[name]['netplan_state']['missing_vrf_link'] = vrf[0]
            if vrf[1]:
                iface[name]['system_state']['missing_vrf_link'] = vrf[1]

        if interfaces != [[], []]:
            system = set(interfaces[0])
            netplan = set(interfaces[1])

            if system != netplan:
                if missing_system := netplan - system:
                    iface[name]['system_state']['missing_interfaces'] = list(missing_system)

                if missing_netplan := system - netplan:
                    iface[name]['netplan_state']['missing_interfaces'] = list(missing_netplan)

    def _normalize_routes(self, routes: set) -> set:
        ''' Apply some transformations to Netplan routes so their representation
        will match the system's.
        '''
        new_routes_set = set()
        for route in routes:
            # If the table is unspecified we set it to main
            if route.table == NetplanRoute._TABLE_UNSPEC_:
                route.table = self._default_route_tables_name_to_number('main')

            # If the addresses are IPv6, compress them so it will match the system representation
            route.to = self._compress_ipv6_address(route.to)
            route.from_addr = self._compress_ipv6_address(route.from_addr)
            route.via = self._compress_ipv6_address(route.via)

            # If the route.to prefix is either /32 and /128 we remove it to match
            # the system representation:
            if route.to != 'default':
                ip_prefix = route.to.split('/')
                if ip_prefix[1] == '32' or ip_prefix[1] == '128':
                    route.to = ip_prefix[0]

            new_routes_set.add(route)

        return new_routes_set

    def _normalize_gateways(self, config: dict) -> AbstractSet[NetplanRoute]:
        ''' Convert the gateway4 and gateway6 properties to NetplanRoutes.
        The only information stored by these properties is the destination IP for
        the default route. Networkd and Network Manager will set a default metric
        for it when the route is installed. As we have no control over it, there
        will always be a diff. The same is true for the route table number.
        To avoid the noise in the diff report when these properties are used,
        we use the following heuristic: if there is one and only one static default
        route in the system and gateway4 and/or gateway6 are used, we set the same
        metric and table as found in the system.
        '''
        routes_set = set()
        system_interfaces = self.system_state.get_data()
        interface = config.get('id')

        ifaces = [iface for iface in system_interfaces.values()
                  if iface.get('id') == interface]
        routes = [route for iface in ifaces for route in iface.get('routes', [])]

        if gateway4 := config.get('gateway4'):
            default_routes = [route for route in routes
                              if route.get('to') == 'default'
                              and route.get('family') == AF_INET.value
                              and route.get('protocol') == 'static']

            route = NetplanRoute(to='default', via=gateway4, family=AF_INET.value, protocol='static')

            if len(default_routes) == 1:
                if metric := default_routes[0].get('metric'):
                    route.metric = metric
                if table := default_routes[0].get('table'):
                    route.table = self._default_route_tables_name_to_number(table)

            routes_set.add(route)

        if gateway6 := config.get('gateway6'):
            default_routes = [route for route in routes
                              if route.get('to') == 'default'
                              and route.get('family') == AF_INET6.value
                              and route.get('protocol') == 'static']

            # Compress the address so it will match the system representation
            gateway6 = self._compress_ipv6_address(gateway6)

            route = NetplanRoute(to='default', via=gateway6, family=AF_INET6.value, protocol='static')

            if len(default_routes) == 1:
                if metric := default_routes[0].get('metric'):
                    route.metric = metric
                if table := default_routes[0].get('table'):
                    route.table = self._default_route_tables_name_to_number(table)

            routes_set.add(route)

        return routes_set

    def _filter_system_routes(self, system_routes: AbstractSet[NetplanRoute], system_addresses: list[str], config: dict) -> set:
        '''
        Some routes found in the system are installed automatically/dynamically without
        being configured in Netplan.
        Here we implement some heuristics to remove these routes from the list we want
        to compare. We do that because these type of routes will probably never be found in the
        Netplan configuration so there is no point in comparing them against Netplan.
        '''

        local_networks = [str(ipaddress.ip_interface(ip).network) for ip in system_addresses]
        # filter out the local link network as we give special treatment to it
        local_networks = list(filter(lambda n: n != 'fe80::/64', local_networks))
        addresses = [str(ipaddress.ip_interface(ip).ip) for ip in system_addresses]
        link_local = config.get('netplan_state', {}).get('link_local', [])
        routes = set()
        for route in system_routes:
            # Filter out link routes (but not link local as we handle them differently)
            if route.scope == 'link' and route.to != 'default' and not ipaddress.ip_interface(route.to).is_link_local:
                continue

            # Filter out routes installed by DHCP
            if route.protocol == 'dhcp':
                continue

            # Filter out routes installed by RA if accept_ra is not false
            accept_ra = config.get('netplan_state', {}).get('accept_ra')
            if route.protocol == 'ra' and accept_ra is not False:
                continue

            # Filter out Link Local routes
            # We only filter them out if the respective 'link-local' setting is present in the netdef
            if route.to != 'default':
                route_to = ipaddress.ip_interface(route.to)
                if route_to.is_link_local:
                    if route.family == AF_INET6.value and 'ipv6' in link_local:
                        continue
                    if route.family == AF_INET.value and 'ipv4' in link_local:
                        continue

            # Filter out host scoped routes
            if (route.scope == 'host' and route.type == 'local' and
                    (route.to in addresses or ipaddress.ip_interface(route.to).is_loopback)):
                continue

            # Filter out the default IPv6 multicast route
            if route.family == AF_INET6.value and route.type == 'multicast' and route.to == 'ff00::/8':
                continue

            # Filter IPv6 local routes
            if (route.family == AF_INET6.value and route.protocol != 'ra'
                    and (route.to in local_networks or route.to in addresses)):
                continue

            routes.add(route)
        return routes

    def _get_netplan_interfaces(self) -> dict:
        system_interfaces = self.system_state.get_data()
        interfaces = {}
        for interface, config in self.netplan_state.netdefs.items():

            iface = {}
            iface[interface] = {'netplan_state': {'id': interface}}
            iface_ref = iface[interface]['netplan_state']

            iface_ref['type'] = DEVICE_TYPES.get(config.type, 'other')

            iface_ref['dhcp4'] = config.dhcp4
            iface_ref['dhcp6'] = config.dhcp6

            iface_ref['link_local'] = config.link_local

            if config.accept_ra is not None:
                iface_ref['accept_ra'] = config.accept_ra

            addresses = [addr for addr in config.addresses]
            if addresses:
                iface_ref['addresses'] = {}
                for addr in addresses:
                    flags = {}
                    if addr.label:
                        flags['label'] = addr.label
                    if addr.lifetime:
                        flags['lifetime'] = addr.lifetime
                    iface_ref['addresses'][str(addr)] = {'flags': flags}

            if nameservers := list(config.nameserver_addresses):
                iface_ref['nameservers_addresses'] = nameservers

            if search := list(config.nameserver_search):
                iface_ref['nameservers_search'] = search

            if routes := list(config.routes):
                iface_ref['routes'] = routes

            if gateway4 := config._gateway4:
                iface_ref['gateway4'] = gateway4

            if gateway6 := config._gateway6:
                iface_ref['gateway6'] = gateway6

            if mac := config.macaddress:
                iface_ref['macaddress'] = mac

            if bridge := config.links.get('bridge'):
                iface_ref['bridge'] = bridge.id

            if bond := config.links.get('bond'):
                iface_ref['bond'] = bond.id

            if vrf := config.links.get('vrf'):
                iface_ref['vrf'] = vrf.id

            if interface not in system_interfaces:
                # If the netdef ID doesn't correspond to any interface name in the system,
                # it might be associated with multiple system interfaces, such as when the 'match' key is used,
                # or the interface name is set in the passthrough section, such as when we create a connection via
                # Network Manager and the netdef ID is the UUID of the connetion.
                # In these cases, we need to look for all the system's interfaces
                # pointing to this netdef and add one netdef entry per device.
                found_some = False
                for key, value in system_interfaces.items():
                    if netdef_id := value.get('id'):
                        if netdef_id == interface:
                            found_some = True
                            interfaces[key] = iface[interface]

                # If we don't find any system interface associated with the netdef
                # that's because it's not matching any device. In this case, we add the
                # netdef ID to the list anyway.
                if not found_some:
                    interfaces.update(iface)
            else:
                interfaces.update(iface)

        self._netplan_state_find_parents(interfaces)
        return interfaces

    def _netplan_state_find_parents(self, interfaces: dict) -> None:
        ''' Associates interfaces with their parents '''
        parents = defaultdict(set)
        for interface, config in interfaces.items():
            if link := config['netplan_state'].get('bridge'):
                parents[link].add(interface)
            if link := config['netplan_state'].get('bond'):
                parents[link].add(interface)
            if link := config['netplan_state'].get('vrf'):
                parents[link].add(interface)

        for interface, members in parents.items():
            interfaces[interface]['netplan_state']['interfaces'] = list(members)

    def _get_system_interfaces(self) -> dict:
        interfaces = {}

        for interface, config in self.system_state.get_data().items():
            if interface == 'netplan-global-state':
                continue

            device_type = config.get('type')
            interfaces[interface] = {'system_state': {'type': device_type}}

            if netdef_id := config.get('id'):
                interfaces[interface]['system_state']['id'] = netdef_id

            iface_ref = interfaces[interface]['system_state']

            if index := config.get('index'):
                iface_ref['index'] = index

            addresses = {}
            for addr in config.get('addresses', []):
                ip = list(addr.keys())[0]
                prefix = addr.get(ip).get('prefix')
                full_addr = f'{ip}/{prefix}'

                addresses[full_addr] = {'flags': addr.get(ip).get('flags', [])}
            if addresses:
                iface_ref['addresses'] = addresses

            if nameservers := config.get('dns_addresses'):
                iface_ref['nameservers_addresses'] = nameservers

            if search := config.get('dns_search'):
                iface_ref['nameservers_search'] = search

            if routes := config.get('routes'):
                iface_ref['routes'] = [self._system_route_to_netplan(route) for route in routes]

            if mac := config.get('macaddress'):
                iface_ref['macaddress'] = mac

            if uplink_interfaces := config.get('interfaces'):
                iface_ref['interfaces'] = uplink_interfaces

            if bond := config.get('bond'):
                iface_ref['bond'] = bond

            if bridge := config.get('bridge'):
                iface_ref['bridge'] = bridge

            if vrf := config.get('vrf'):
                iface_ref['vrf'] = vrf

        return interfaces

    def _system_route_to_netplan(self, system_route: dict) -> NetplanRoute:
        route = {}

        if family := system_route.get('family'):
            route['family'] = family
        if to := system_route.get('to'):
            route['to'] = to
        if via := system_route.get('via'):
            route['via'] = via
        if from_addr := system_route.get('from'):
            route['from_addr'] = from_addr
        if metric := system_route.get('metric'):
            route['metric'] = metric
        if scope := system_route.get('scope'):
            route['scope'] = scope
        if route_type := system_route.get('type'):
            route['type'] = route_type
        if protocol := system_route.get('protocol'):
            route['protocol'] = protocol
        if table := system_route.get('table'):
            route['table'] = self._default_route_tables_name_to_number(table)

        return NetplanRoute(**route)

    def _default_route_tables_name_to_number(self, name: str) -> int:
        if name.isdigit():
            return int(name)

        if not self.route_lookup_table_names:
            self.route_lookup_table_names = route_table_lookup()

        return self.route_lookup_table_names.get(name, 0)



# Copyright (C) 2023 Canonical, Ltd.
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import IntEnum
from typing import Union, IO

from ._netplan_cffi import ffi, lib
from ._utils import _checked_lib_call


class Flags(IntEnum):
    IGNORE_ERRORS = 1 << 0


class Parser():
    def __init__(self):
        self._ptr = lib.netplan_parser_new()

    def __del__(self):
        ref = ffi.new('NetplanParser **', self._ptr)
        lib.netplan_parser_clear(ref)

    def load_yaml(self, input_file: Union[str, IO]):
        if isinstance(input_file, str):
            return _checked_lib_call(lib.netplan_parser_load_yaml, self._ptr, input_file.encode('utf-8'))
        else:
            return _checked_lib_call(lib.netplan_parser_load_yaml_from_fd, self._ptr, input_file.fileno())

    def load_yaml_hierarchy(self, rootdir: str = None):
        root = rootdir.encode('utf-8') if rootdir else ffi.NULL
        return _checked_lib_call(lib.netplan_parser_load_yaml_hierarchy, self._ptr, root)

    def load_keyfile(self, input_file: str):  # TODO: load from File/fd (i.e. input_file: Union[str, IO])
        return _checked_lib_call(lib.netplan_parser_load_keyfile, self._ptr, input_file.encode('utf-8'))

    def load_nullable_fields(self, input_file: IO):
        return _checked_lib_call(lib.netplan_parser_load_nullable_fields, self._ptr, input_file.fileno())

    def _load_nullable_overrides(self, input_file: IO, constraint: str):
        return _checked_lib_call(lib.netplan_parser_load_nullable_overrides,
                                 self._ptr, input_file.fileno(), constraint.encode('utf-8'))

    @property
    def flags(self) -> int:
        return lib.netplan_parser_get_flags(self._ptr)

    @flags.setter
    def flags(self, flags: int):
        _ = _checked_lib_call(lib.netplan_parser_set_flags, self._ptr, flags)

    @property
    def error_count(self) -> int:
        return lib.netplan_parser_get_error_count(self._ptr)



#!/usr/bin/python3
#
# Copyright (C) 2022 Canonical, Ltd.
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''netplan status command line'''

import json
import logging
import re
from netplan.netdef import NetplanRoute

import yaml

from .. import utils
from ..state import NetplanConfigState, SystemConfigState, JSON
from ..state_diff import DiffJSONEncoder, NetplanDiffState


MATCH_TAGS = re.compile(r'\[([a-z0-9]+)\].*\[\/\1\]')
RICH_OUTPUT = False
try:
    from rich.console import Console
    from rich.highlighter import RegexHighlighter
    from rich.theme import Theme

    class NetplanHighlighter(RegexHighlighter):
        base_style = 'netplan.'
        highlights = [
            r'(^|[\s\/])(?P<int>\d+)([\s:]?\s|$)',
            r'(?P<str>(\"|\').+(\"|\'))',
            ]
    RICH_OUTPUT = True
except ImportError:  # pragma: nocover (we mock RICH_OUTPUT, ignore the logging)
    logging.debug("python3-rich not found, falling back to plain output")


class NetplanStatus(utils.NetplanCommand):
    def __init__(self):
        super().__init__(command_id='status',
                         description='Query networking state of the running system',
                         leaf=True)
        self.all = False
        self.state_diff = None
        self.route_lookup_table_names = {}

    def run(self):
        self.parser.add_argument('ifname', nargs='?', type=str, default=None,
                                 help='Show only this interface')
        self.parser.add_argument('-a', '--all', action='store_true',
                                 help='Show all interface data (incl. inactive)')
        self.parser.add_argument('-v', '--verbose', action='store_true',
                                 help='Show extra information')
        self.parser.add_argument('-f', '--format', default='tabular',
                                 help='Output in machine readable `json` or `yaml` format')
        self.parser.add_argument('--diff', action='store_true',
                                 help='Show the differences between the system\'s and netplan\'s states')
        self.parser.add_argument('--diff-only', action='store_true',
                                 help='Only show the differences between the system\'s and netplan\'s states')
        self.parser.add_argument('--root-dir',
                                 help='Search for configuration files in this root directory instead of /')

        self.func = self.command
        self.parse_args()
        self.run_command()

    def _create_pretty_print(self, _console_width):
        if RICH_OUTPUT:
            # TODO: Use a proper (subiquity?) color palette
            theme = Theme({
                'netplan.int': 'bold cyan',
                'netplan.str': 'yellow',
                'muted': 'grey62',
                'online': 'green bold',
                'offline': 'red bold',
                'unknown': 'yellow bold',
                'highlight': 'bold'
                })
            if self.diff:
                theme = Theme({
                    'netplan.int': 'grey62',
                    'netplan.str': 'grey62',
                    'muted': 'grey62',
                    'online': 'green bold',
                    'offline': 'red bold',
                    'unknown': 'yellow bold',
                    'highlight': 'bold'
                    })

            console = Console(highlighter=NetplanHighlighter(), theme=theme,
                              width=_console_width, emoji=False)
            pprint = console.print
        else:
            pprint = self.plain_print

        return pprint

    def _get_interface_diff(self, ifname) -> dict:
        if self.state_diff:
            if diff := self.state_diff['interfaces'].get(ifname):
                if diff.get('system_state') or diff.get('netplan_state'):
                    return diff
        return {}

    def _is_interface_missing_in_netplan(self, ifname) -> bool:
        if self.state_diff:
            if missing := self.state_diff.get('missing_interfaces_netplan'):
                if ifname in missing:
                    return True
        return False

    def _get_missing_property_list(self, ifname: str, state: str, property: str) -> list[str]:
        if self.state_diff:
            if diff := self.state_diff['interfaces'].get(ifname):
                return diff.get(state, {}).get(property, [])
        return []

    def _get_missing_property_str(self, ifname: str, state: str, property: str) -> str:
        if self.state_diff:
            if diff := self.state_diff['interfaces'].get(ifname):
                return diff.get(state, {}).get(property, '')
        return ''

    def _get_missing_property_set(self, ifname: str, state: str, property: str) -> set:
        if self.state_diff:
            if diff := self.state_diff['interfaces'].get(ifname):
                return diff.get(state, {}).get(property, set())
        return set()

    def _get_missing_property_bool(self, ifname: str, state: str, property: str) -> bool:
        if self.state_diff:
            if diff := self.state_diff['interfaces'].get(ifname):
                return diff.get(state, {}).get(property, False)
        return False

    def _get_missing_netplan_addresses(self, ifname) -> list[str]:
        return self._get_missing_property_list(ifname, 'netplan_state', 'missing_addresses')

    def _get_missing_system_nameservers(self, ifname) -> list[str]:
        return self._get_missing_property_list(ifname, 'system_state', 'missing_nameservers_addresses')

    def _get_missing_netplan_nameservers(self, ifname) -> list[str]:
        return self._get_missing_property_list(ifname, 'netplan_state', 'missing_nameservers_addresses')

    def _get_missing_netplan_search(self, ifname) -> list[str]:
        return self._get_missing_property_list(ifname, 'netplan_state', 'missing_nameservers_search')

    def _get_missing_system_search(self, ifname) -> list[str]:
        return self._get_missing_property_list(ifname, 'system_state', 'missing_nameservers_search')

    def _get_missing_system_macaddress(self, ifname) -> str:
        return self._get_missing_property_str(ifname, 'system_state', 'missing_macaddress')

    def _get_missing_netplan_routes(self, ifname) -> set[NetplanRoute]:
        return self._get_missing_property_set(ifname, 'netplan_state', 'missing_routes')

    def _get_missing_system_routes(self, ifname) -> set[NetplanRoute]:
        return self._get_missing_property_set(ifname, 'system_state', 'missing_routes')

    def _is_missing_dhcp4_address(self, ifname) -> bool:
        return self._get_missing_property_bool(ifname, 'system_state', 'missing_dhcp4_address')

    def _is_missing_dhcp6_address(self, ifname) -> bool:
        return self._get_missing_property_bool(ifname, 'system_state', 'missing_dhcp6_address')

    def _get_missing_system_bond_link(self, ifname) -> str:
        return self._get_missing_property_str(ifname, 'system_state', 'missing_bond_link')

    def _get_missing_netplan_bond_link(self, ifname) -> str:
        return self._get_missing_property_str(ifname, 'netplan_state', 'missing_bond_link')

    def _get_missing_system_bridge_link(self, ifname) -> str:
        return self._get_missing_property_str(ifname, 'system_state', 'missing_bridge_link')

    def _get_missing_netplan_bridge_link(self, ifname) -> str:
        return self._get_missing_property_str(ifname, 'netplan_state', 'missing_bridge_link')

    def _get_missing_system_vrf_link(self, ifname) -> str:
        return self._get_missing_property_str(ifname, 'system_state', 'missing_vrf_link')

    def _get_missing_netplan_vrf_link(self, ifname) -> str:
        return self._get_missing_property_str(ifname, 'netplan_state', 'missing_vrf_link')

    def _get_missing_netplan_members(self, ifname) -> list[str]:
        return self._get_missing_property_list(ifname, 'netplan_state', 'missing_interfaces')

    def _get_missing_system_members(self, ifname) -> list[str]:
        return self._get_missing_property_list(ifname, 'system_state', 'missing_interfaces')

    def _get_missing_system_interfaces(self) -> dict:
        if self.state_diff:
            return self.state_diff.get('missing_interfaces_system', {})
        return {}

    def _has_diff(self, ifname) -> bool:
        if self._is_interface_missing_in_netplan(ifname):
            return True
        if self.state_diff:
            if diff := self.state_diff['interfaces'].get(ifname):
                if diff.get('system_state') or diff.get('netplan_state'):
                    return True
        return False

    def _display_global_state(self, data):
        global_state = data.get('netplan-global-state', {})
        self.pprint(('{title:>'+self.PAD+'} {value}').format(
            title='Online state:',
            value='[online]online[/online]' if global_state.get('online', False) else '[offline]offline[/offline]',
            ))
        ns = global_state.get('nameservers', {})
        dns_addr: list = ns.get('addresses', [])
        dns_mode: str = ns.get('mode')
        dns_search: list = ns.get('search', [])
        if dns_addr:
            for i, val in enumerate(dns_addr):
                self.pprint(('{title:>'+self.PAD+'} {value}[muted]{mode}[/muted]').format(
                    title='DNS Addresses:' if i == 0 else '',
                    value=val,
                    mode=' ({})'.format(dns_mode) if dns_mode else '',
                    ))
        if dns_search:
            for i, val in enumerate(dns_search):
                self.pprint(('{title:>'+self.PAD+'} {value}').format(
                    title='DNS Search:' if i == 0 else '',
                    value=val,
                    ))
        self.pprint()

    def _display_interface_header(self, ifname: str, data):
        state = data.get('operstate', 'UNKNOWN') + '/' + data.get('adminstate', 'UNKNOWN')
        scolor = 'unknown'
        if state == 'UP/UP':
            state = 'UP'
            scolor = 'online'
        elif state == 'DOWN/DOWN':
            state = 'DOWN'
            scolor = 'offline'
        full_type = data.get('type', 'other')
        ssid = data.get('ssid')
        tunnel_mode = data.get('tunnel_mode')
        if full_type == 'wifi' and ssid:
            full_type += ('/"' + ssid + '"')
        elif full_type == 'tunnel' and tunnel_mode:
            full_type += ('/' + tunnel_mode)

        format = '[{col}]●[/{col}] {idx:>2}: {name} {type} [{col}]{state}[/{col}] ({backend}{netdef})'
        netdef = ': [highlight]{}[/highlight]'.format(data.get('id')) if data.get('id') else ''
        extra = ''
        sign = ''
        if self.diff:
            if self._is_interface_missing_in_netplan(ifname):
                sign = self.PLUS
                format = '{sign} [{col}]●[/{col}] {idx:>2}: [green][highlight]{name} {type}'
                format += ' [{col}]{state}[/{col}] ({backend}{netdef})[/highlight][/green]'
            else:
                format = '  [muted]● {idx:>2}: {name} {type} {state} ({backend}{netdef})[/muted]'
                netdef = ': {}'.format(data.get('id')) if data.get('id') else ''

        if not self.diff_only or self._has_diff(ifname):
            self.pprint(format.format(
                sign=sign,
                col=scolor,
                idx=data.get('index', '?'),
                name=ifname,
                type=full_type,
                state=state,
                backend=data.get('backend', 'unmanaged'),
                netdef=netdef,
                extra=extra,
                ))

    def _display_mac_address(self, ifname: str, data):
        if macaddress := data.get('macaddress'):
            hide_macaddress = False
            missing_system_macaddress = self._get_missing_system_macaddress(ifname)
            format = '{title:>'+self.PAD+'} {mac}[muted]{vendor}[/muted]'
            sign = ''
            if self.diff and not missing_system_macaddress:
                format = '  [muted]{title:>'+self.PAD+'} {mac}{vendor}[/muted]'
                if self.diff_only:
                    hide_macaddress = True
            elif self.diff and missing_system_macaddress:
                sign = self.PLUS
                format = '{sign} {title:>'+self.PAD+'} [green][highlight]{mac}{vendor}[/highlight][/green]'

            if not hide_macaddress:
                self.pprint((format).format(
                    sign=sign,
                    title='MAC Address:',
                    mac=macaddress,
                    vendor=' ({})'.format(data.get('vendor', '')) if data.get('vendor') else '',
                    ))

                if self.diff and missing_system_macaddress:
                    sign = self.MINUS
                    format = '{sign} {title:>'+self.PAD+'} [red][highlight]{mac}{vendor}[/highlight][/red]'
                    self.pprint((format).format(
                        sign=sign,
                        title='',
                        mac=missing_system_macaddress,
                        vendor=' ({})'.format(data.get('vendor', '')) if data.get('vendor') else '',
                        ))

    def _display_ip_addresses(self, ifname: str, data):
        lst: list = data.get('addresses', [])
        addresses_displayed = 0
        if lst:
            missing_netplan_addresses = self._get_missing_netplan_addresses(ifname)
            for obj in lst:
                sign = ''
                hide_address = False
                ip, extra = list(obj.items())[0]  # get first (and only) address
                prefix = extra.get('prefix', '')
                flags = []
                if extra.get('flags'):  # flags
                    flags = extra.get('flags', [])
                highlight_start = ''
                highlight_end = ''
                if not flags or 'dhcp' in flags:
                    highlight_start = '[highlight]'
                    highlight_end = '[/highlight]'

                address = f'{ip}/{prefix}'
                if self.diff and address not in missing_netplan_addresses:
                    format = '  [muted]{title:>'+self.PAD+'} {start}{ip}/{prefix}{end}{extra}[/muted]'
                    highlight_start = ''
                    highlight_end = ''
                    if self.diff_only:
                        hide_address = True
                elif self.diff and address in missing_netplan_addresses:
                    sign = self.PLUS
                    format = '{sign} {title:>'+self.PAD+'} [green]{start}{ip}/{prefix}{extra}{end}[/green]'
                    highlight_start = '[highlight]'
                    highlight_end = '[/highlight]'
                else:
                    format = '{title:>'+self.PAD+'} {start}{ip}/{prefix}{end}[muted]{extra}[/muted]'

                if not hide_address:
                    self.pprint((format).format(
                        sign=sign,
                        title='Addresses:' if addresses_displayed == 0 else '',
                        ip=ip,
                        prefix=prefix,
                        extra=' ('+', '.join(flags)+')' if flags else '',
                        start=highlight_start,
                        end=highlight_end,
                        ))
                    addresses_displayed += 1

        if diff := self._get_interface_diff(ifname):
            sign = self.MINUS
            if missing_addresses := diff.get('system_state', {}).get('missing_addresses'):
                for ip in missing_addresses:
                    self.pprint(('{sign} {title:>'+self.PAD+'} [highlight][red]{ip}[/red][/highlight]').format(
                        sign=sign,
                        title='Addresses:' if addresses_displayed == 0 else '',
                        ip=ip,
                        ))
                    addresses_displayed += 1
            if self._is_missing_dhcp4_address(ifname):
                self.pprint(('{sign} {title:>'+self.PAD+'} [highlight][red]0.0.0.0/0 (dhcp)[/red][/highlight]').format(
                    sign=sign,
                    title='Addresses:' if addresses_displayed == 0 else '',
                    ))
                addresses_displayed += 1
            if self._is_missing_dhcp6_address(ifname):
                self.pprint(('{sign} {title:>'+self.PAD+'} [highlight][red]::/0 (dhcp)[/red][/highlight]').format(
                    sign=sign,
                    title='Addresses:' if addresses_displayed == 0 else '',
                    ))

    def _display_dns_addresses(self, ifname: str, data):
        lst = data.get('dns_addresses', [])
        nameservers_displayed = 0
        if lst:
            missing_netplan_nameservers = self._get_missing_netplan_nameservers(ifname)
            for val in lst:
                sign = ''
                hide_nameserver = False
                if self.diff and val not in missing_netplan_nameservers:
                    format = '  [muted]{title:>'+self.PAD+'} {value}[/muted]'
                    highlight_start = ''
                    highlight_end = ''
                    if self.diff_only:
                        hide_nameserver = True
                elif self.diff and val in missing_netplan_nameservers:
                    sign = self.PLUS
                    format = '{sign} {title:>'+self.PAD+'} [green]{start}{value}{end}[/green]'
                    highlight_start = '[highlight]'
                    highlight_end = '[/highlight]'
                else:
                    format = '{title:>'+self.PAD+'} {value}'
                    highlight_start = ''
                    highlight_end = ''

                if not hide_nameserver:
                    self.pprint((format).format(
                        sign=sign,
                        title='DNS Addresses:' if nameservers_displayed == 0 else '',
                        value=val,
                        start=highlight_start,
                        end=highlight_end
                        ))
                    nameservers_displayed += 1

        if self._has_diff(ifname):
            if missing_nameservers_addresses := self._get_missing_system_nameservers(ifname):
                sign = self.MINUS
                for ip in missing_nameservers_addresses:
                    self.pprint(('{sign} {title:>'+self.PAD+'} [red][highlight]{ip}[/highlight][/red]').format(
                        sign=sign,
                        title='DNS Addresses:' if nameservers_displayed == 0 else '',
                        ip=ip,
                        ))
                    nameservers_displayed += 1

    def _display_dns_search(self, ifname, data):
        lst = data.get('dns_search', [])
        searches_displayed = 0
        if lst:
            missing_netplan_search = self._get_missing_netplan_search(ifname)
            for i, val in enumerate(lst):
                sign = ''
                hide_search = False
                if self.diff and val not in missing_netplan_search:
                    format = '  [muted]{title:>'+self.PAD+'} {value}[/muted]'
                    highlight_start = ''
                    highlight_end = ''
                    if self.diff_only:
                        hide_search = True
                elif self.diff and val in missing_netplan_search:
                    sign = self.PLUS
                    format = '{sign} {title:>'+self.PAD+'} [green]{start}{value}{end}[/green]'
                    highlight_start = '[highlight]'
                    highlight_end = '[/highlight]'
                else:
                    format = '{title:>'+self.PAD+'} {value}'
                    highlight_start = ''
                    highlight_end = ''

                if not hide_search:
                    self.pprint((format).format(
                        sign=sign,
                        title='DNS Search:' if searches_displayed == 0 else '',
                        value=val,
                        start=highlight_start,
                        end=highlight_end
                        ))
                    searches_displayed += 1

        if self._has_diff(ifname):
            if missing_nameservers_search := self._get_missing_system_search(ifname):
                sign = self.MINUS
                for domain in missing_nameservers_search:
                    self.pprint(('{sign} {title:>'+self.PAD+'} [red][highlight]{domain}[/highlight][/red]').format(
                        sign=sign,
                        title='DNS Search:' if searches_displayed == 0 else '',
                        domain=domain,
                        ))
                    searches_displayed += 1

    def _display_routes(self, ifname, data):
        lst = data.get('routes', [])
        missing_netplan_routes = self._get_missing_netplan_routes(ifname)
        missing_system_routes = self._get_missing_system_routes(ifname)
        routes_displayed = 0
        if lst:
            if not self.route_lookup_table_names:
                self.route_lookup_table_names = utils.route_table_lookup()

            diff_state = NetplanDiffState(None, None)
            routes = [diff_state._system_route_to_netplan(route) for route in lst]
            if not self.verbose:
                # filter out routes that are not in the main route table
                routes = filter(lambda r: r.table == 254, routes)
            for route in routes:
                hide_route = False
                default_start = ''
                default_end = ''
                if route.to == 'default':
                    default_start = '[highlight]'
                    default_end = '[/highlight]'
                via = ''
                if route.via:
                    via = ' via ' + route.via
                src = ''
                if route.from_addr:
                    src = ' from ' + route.from_addr
                metric = ''
                if route.metric < NetplanRoute._METRIC_UNSPEC_:
                    metric = ' metric ' + str(route.metric)
                table = ''
                if self.verbose and route.table > 0:
                    table = ' table {}'.format(self.route_lookup_table_names.get(route.table, route.table))

                extra = []
                if route.protocol and route.protocol != 'kernel':
                    proto = route.protocol
                    extra.append(proto)
                if route.scope and route.scope != 'global':
                    scope = route.scope
                    extra.append(scope)
                if route.type and route.type != 'unicast':
                    type = route.type
                    extra.append(type)

                sign = ''
                if self.diff and route not in missing_netplan_routes:
                    format = '  [muted]{title:>'+self.PAD+'} {start}{to}{via}{src}{metric}{table}{end}{extra}[/muted]'
                    default_start = ''
                    default_end = ''
                    if self.diff_only:
                        hide_route = True
                elif self.diff and route in missing_netplan_routes:
                    sign = self.PLUS
                    format = '{sign} {title:>'+self.PAD+'} [green][highlight]{start}{to}{via}{src}{metric}'
                    format += '{table}{end}{extra}[/highlight][/green]'
                else:
                    format = '{title:>'+self.PAD+'} {start}{to}{via}{src}{metric}{table}{end}[muted]{extra}[/muted]'

                if not hide_route:
                    self.pprint(format.format(
                        sign=sign,
                        title='Routes:' if routes_displayed == 0 else '',
                        to=route.to,
                        via=via,
                        src=src,
                        metric=metric,
                        table=table,
                        extra=' ('+', '.join(extra)+')' if extra else '',
                        start=default_start,
                        end=default_end))
                    routes_displayed += 1

        if self.diff:
            for route in missing_system_routes:
                via = ''
                if route.via:
                    via = ' via ' + route.via
                src = ''
                if route.from_addr:
                    src = ' from ' + route.from_addr
                metric = ''
                if route.metric < NetplanRoute._METRIC_UNSPEC_:
                    metric = ' metric ' + str(route.metric)
                table = ''
                if self.verbose and route.table > 0:
                    table = ' table {}'.format(self.route_lookup_table_names.get(route.table, route.table))

                extra = []
                if route.scope and route.scope != 'global':
                    scope = route.scope
                    extra.append(scope)
                if route.type and route.type != 'unicast':
                    type = route.type
                    extra.append(type)

                sign = self.MINUS
                format = '{sign} {title:>'+self.PAD+'} {start}[red]{to}{via}{src}{metric}{table}{extra}[/red]{end}'
                self.pprint(format.format(
                    sign=sign,
                    title='Routes:' if routes_displayed == 0 else '',
                    to=route.to,
                    via=via,
                    src=src,
                    metric=metric,
                    table=table,
                    extra=' ('+', '.join(extra)+')' if extra else '',
                    start='[highlight]',
                    end='[/highlight]'))
                routes_displayed += 1

    def _display_bridge(self, ifname, data):
        val = data.get('bridge')
        if val:
            missing_netplan_bridge_link = self._get_missing_netplan_bridge_link(ifname)
            format = '{title:>'+self.PAD+'} {value}'
            sign = ''
            hide_bridge = False
            if self.diff and not missing_netplan_bridge_link:
                format = '  [muted]{title:>'+self.PAD+'} {value}[/muted]'
                if self.diff_only:
                    hide_bridge = True
            elif self.diff and missing_netplan_bridge_link:
                sign = self.PLUS
                format = '{sign} {title:>'+self.PAD+'} [highlight][green]{value}[/green][/highlight]'
                val = missing_netplan_bridge_link

            if not hide_bridge:
                self.pprint((format).format(
                    sign=sign,
                    title='Bridge:',
                    value=val,
                    ))
        if missing_system_bridge_link := self._get_missing_system_bridge_link(ifname):
            sign = self.MINUS
            format = '{sign} {title:>'+self.PAD+'} [highlight][red]{value}[/red][/highlight]'
            self.pprint((format).format(
                sign=sign,
                title='Bridge:',
                value=missing_system_bridge_link,
            ))

    def _display_bond(self, ifname, data):
        val = data.get('bond')
        if val:
            missing_netplan_bond_link = self._get_missing_netplan_bond_link(ifname)
            format = '{title:>'+self.PAD+'} {value}'
            sign = ''
            hide_bond = False
            if self.diff and not missing_netplan_bond_link:
                format = '  [muted]{title:>'+self.PAD+'} {value}[/muted]'
                if self.diff_only:
                    hide_bond = True
            elif self.diff and missing_netplan_bond_link:
                sign = self.PLUS
                format = '{sign} {title:>'+self.PAD+'} [highlight][green]{value}[/green][/highlight]'
                val = missing_netplan_bond_link

            if not hide_bond:
                self.pprint((format).format(
                    sign=sign,
                    title='Bond:',
                    value=val,
                    ))
        if missing_system_bond_link := self._get_missing_system_bond_link(ifname):
            sign = self.MINUS
            format = '{sign} {title:>'+self.PAD+'} [highlight][red]{value}[/red][/highlight]'
            self.pprint((format).format(
                sign=sign,
                title='Bond:',
                value=missing_system_bond_link,
            ))

    def _display_vrf(self, ifname, data):
        val = data.get('vrf')
        if val:
            missing_netplan_vrf_link = self._get_missing_netplan_vrf_link(ifname)
            format = '{title:>'+self.PAD+'} {value}'
            sign = ''
            hide_vrf = False
            if self.diff and not missing_netplan_vrf_link:
                format = '  [muted]{title:>'+self.PAD+'} {value}[/muted]'
                if self.diff_only:
                    hide_vrf = True
            elif self.diff and missing_netplan_vrf_link:
                sign = self.PLUS
                format = '{sign} {title:>'+self.PAD+'} [highlight][green]{value}[/green][/highlight]'
                val = missing_netplan_vrf_link

            if not hide_vrf:
                self.pprint((format).format(
                    sign=sign,
                    title='VRF:',
                    value=val,
                    ))
        if missing_system_vrf_link := self._get_missing_system_vrf_link(ifname):
            sign = self.MINUS
            format = '{sign} {title:>'+self.PAD+'} [highlight][red]{value}[/red][/highlight]'
            self.pprint((format).format(
                sign=sign,
                title='VRF:',
                value=missing_system_vrf_link,
            ))

    def _display_members(self, ifname: str, data):
        lst = data.get('interfaces', [])
        members_displayed = 0
        if lst:
            missing_netplan_interfaces = self._get_missing_netplan_members(ifname)
            for val in lst:
                sign = ''
                hide_member = False
                if self.diff and val not in missing_netplan_interfaces:
                    format = '  [muted]{title:>'+self.PAD+'} {value}[/muted]'
                    highlight_start = ''
                    highlight_end = ''
                    if self.diff_only:
                        hide_member = True
                elif self.diff and val in missing_netplan_interfaces:
                    sign = self.PLUS
                    format = '{sign} {title:>'+self.PAD+'} [green]{start}{value}{end}[/green]'
                    highlight_start = '[highlight]'
                    highlight_end = '[/highlight]'
                else:
                    format = '{title:>'+self.PAD+'} {value}'
                    highlight_start = ''
                    highlight_end = ''

                if not hide_member:
                    self.pprint((format).format(
                        sign=sign,
                        title='Interfaces:' if members_displayed == 0 else '',
                        value=val,
                        start=highlight_start,
                        end=highlight_end,
                        ))
                    members_displayed += 1

        if self._has_diff(ifname):
            if missing_members := self._get_missing_system_members(ifname):
                sign = self.MINUS
                for member in missing_members:
                    self.pprint(('{sign} {title:>'+self.PAD+'} [red][highlight]{member}[/highlight][/red]').format(
                        sign=sign,
                        title='Interfaces:' if members_displayed == 0 else '',
                        member=member,
                        ))
                    members_displayed += 1

    def _display_activation_mode(self, data):
        val = data.get('activation_mode')
        if val:
            self.pprint(('{title:>'+self.PAD+'} {value}').format(
                title='Activation Mode:',
                value=val,
                ))

    def _display_missing_interfaces(self):
        missing_interfaces = self._get_missing_system_interfaces()
        sign = self.MINUS
        for index, (interface, properties) in enumerate(missing_interfaces.items(), 1):
            # If we called netplan status for a single interface, ignore the rest
            if self.ifname and self.ifname != interface:
                continue

            self.pprint('{sign} [{col}]● {idx:>2}  {name} {type}[/{col}]'.format(
                sign=sign,
                col='red',
                idx='',
                name=interface,
                type=properties.get('type'),
                ))

            if index != len(missing_interfaces) and not self.ifname:
                # linebreak only if it's not the last interfaces or the only one
                self.pprint()

    def plain_print(self, *args, **kwargs):
        if len(args):
            lst = list(args)
            while True:
                tags = MATCH_TAGS.findall(lst[0])
                if not tags:
                    break
                for tag in tags:
                    # remove matching opening and closing tag
                    lst[0] = lst[0].replace('[{}]'.format(tag), '')\
                                   .replace('[/{}]'.format(tag), '')
            return print(*lst, **kwargs)
        return print(*args, **kwargs)

    def pretty_print(self, data: JSON, total: int, _console_width=None) -> None:
        self.pprint = self._create_pretty_print(_console_width)
        self.PLUS = '[green]+[/green]'
        self.MINUS = '[red]-[/red]'
        self.PAD = '18'
        if self.diff:
            # In diff mode we shift the text 2 columns to the right so we can display
            # + and - and maintain the alignment consistency
            self.PAD = '20'

        # Global state
        if not self.diff:
            self._display_global_state(data)

        # Per interface
        interfaces = [(key, data[key]) for key in data if key != 'netplan-global-state']
        if self.diff_only:
            # in diff-only mode we filter out interfaces that don't have any diff
            interfaces = list(filter(lambda i: self._has_diff(i[0]), interfaces))

        missing_interfaces = self._get_missing_system_interfaces()
        for index, (ifname, ifconfig) in enumerate(interfaces, 1):
            # If we called netplan status for a single interface, ignore the rest
            if self.ifname and self.ifname != ifname:
                continue

            self._display_interface_header(ifname, ifconfig)
            self._display_mac_address(ifname, ifconfig)
            self._display_ip_addresses(ifname, ifconfig)
            self._display_dns_addresses(ifname, ifconfig)
            self._display_dns_search(ifname, ifconfig)
            self._display_routes(ifname, ifconfig)
            self._display_bridge(ifname, ifconfig)
            self._display_bond(ifname, ifconfig)
            self._display_vrf(ifname, ifconfig)
            self._display_members(ifname, ifconfig)
            self._display_activation_mode(ifconfig)

            if not self.diff_only or self._has_diff(ifname):
                # we only break to a new line if we still have data to display
                if (index != len(interfaces) or len(missing_interfaces) > 0) and not self.ifname:
                    self.pprint()

        if self.diff:
            self._display_missing_interfaces()

        hidden = total - len(interfaces)
        if hidden > 0 and not self.diff:
            self.pprint('\n{} inactive interfaces hidden. Use "--all" to show all.'.format(hidden))

        if self.diff and not self.diff_only:
            self.pprint(
                '\nUse [yellow]"--diff-only"[/yellow] to omit the information that is consistent between the system and Netplan.'
            )

    def command(self):
        # --diff-only implies --diff
        if self.diff_only:
            self.diff = True

        # --diff needs data from all interfaces to work
        if self.diff:
            self.all = True

        system_state = SystemConfigState(self.ifname, self.all)

        output_format = self.format.lower()

        if self.diff:
            netplan_state = NetplanConfigState(rootdir=self.root_dir)
            diff_state = NetplanDiffState(system_state, netplan_state)

            self.state_diff = diff_state.get_diff(self.ifname)

            if output_format == 'json':
                print(json.dumps(self.state_diff, cls=DiffJSONEncoder))
                return
            elif output_format == 'yaml':
                serialized = json.dumps(self.state_diff, cls=DiffJSONEncoder)
                print(yaml.dump(json.loads(serialized)))
                return

        if output_format == 'json':  # structural JSON output
            print(json.dumps(system_state.get_data()))
        elif output_format == 'yaml':  # stuctural YAML output
            print(yaml.dump(system_state.get_data()))
        else:  # pretty print, human readable output
            self.pretty_print(system_state.get_data(), system_state.number_of_interfaces)



#!/usr/bin/python3
#
# Copyright (C) 2020 Canonical, Ltd.
# Author: Lukas 'slyon' Märdian <lukas.maerdian@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess

OPENVSWITCH_OVS_VSCTL = 'ovs-vsctl'
# Defaults for non-optional settings, as defined here:
# http://www.openvswitch.org/ovs-vswitchd.conf.db.5.pdf
DEFAULTS = {
    # Mandatory columns:
    'mcast_snooping_enable': 'false',
    'rstp_enable': 'false',
}
GLOBALS = {
    # Global commands:
    'set-ssl': ('del-ssl', 'get-ssl'),
    'set-fail-mode': ('del-fail-mode', 'get-fail-mode'),
    'set-controller': ('del-controller', 'get-controller'),
}


class OvsDbServerNotRunning(Exception):
    pass


class OvsDbServerNotInstalled(Exception):
    pass


def _ovs_vsctl_path():  # pragma: nocover
    if os.path.exists('/usr/bin/'+OPENVSWITCH_OVS_VSCTL) or \
            not os.path.exists('/snap/bin/'+OPENVSWITCH_OVS_VSCTL):
        return '/usr/bin/'+OPENVSWITCH_OVS_VSCTL
    else:
        return '/snap/bin/'+OPENVSWITCH_OVS_VSCTL


OVS_VSCTL_PATH = _ovs_vsctl_path()


def _ovs_installed():  # pragma: nocover
    return os.path.exists(OVS_VSCTL_PATH)


def _ovs_active():  # pragma: nocover
    return subprocess.call([OVS_VSCTL_PATH, 'show'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0


def _del_col(type, iface, column, value):
    """Cleanup values from a column (i.e. "column=value")"""
    default = DEFAULTS.get(column)
    if default is None:
        # removes the exact value only if it was set by netplan
        cmd = [OVS_VSCTL_PATH, 'remove', type, iface, column, value]
        logging.debug('Running: %s' % ' '.join(cmd))
        subprocess.check_call(cmd)
    elif default and default != value:
        # reset to default, if its not the default already
        cmd = [OVS_VSCTL_PATH, 'set', type, iface, '%s=%s' % (column, default)]
        logging.debug('Running: %s' % ' '.join(cmd))
        subprocess.check_call(cmd)


def _del_dict(type, iface, column, key, value):
    """Cleanup values from a dictionary (i.e. "column:key=value")"""
    # removes the exact value only if it was set by netplan
    cmd = [OVS_VSCTL_PATH, 'remove', type, iface, column, '%s=\"%s\"' % (key, value)]
    logging.debug('Running: %s' % ' '.join(cmd))
    subprocess.check_call(cmd)


def _del_global(type, iface, key, value):
    """Cleanup commands from the global namespace"""
    del_cmd, get_cmd = GLOBALS.get(key, (None, None))
    if del_cmd == 'del-ssl':
        iface = None

    if del_cmd:
        args_get = [OVS_VSCTL_PATH, get_cmd]
        args_del = [OVS_VSCTL_PATH, del_cmd]
        if iface:
            args_get.append(iface)
            args_del.append(iface)
        # Check the current value of a global command and compare it to the tag-value, e.g.:
        # * get-ssl: netplan/global/set-ssl=/private/key.pem,/another/cert.pem,/some/ca-cert.pem
        # Private key: /private/key.pem
        # Certificate: /another/cert.pem
        # CA Certificate: /some/ca-cert.pem
        # Bootstrap: false
        # * get-fail-mode: netplan/global/set-fail-mode=secure
        # secure
        # * get-controller: netplan/global/set-controller=tcp:127.0.0.1:1337,unix:/some/socket
        # tcp:127.0.0.1:1337
        # unix:/some/socket
        out = subprocess.check_output(args_get, text=True)
        # Clean it only if the exact same value(s) were set by netplan.
        # Don't touch it if other values were set by another integration.
        if all(item in out for item in value.split(',')):
            subprocess.check_call(args_del)
    else:
        raise Exception('Reset command unknown for:', key)


def clear_setting(type, iface, setting, value):
    """Check if this setting is in a dict or a colum and delete accordingly"""
    split = setting.split('/', 2)
    col = split[1]
    if col == 'global' and len(split) > 2:
        _del_global(type, iface, split[2], value)
    elif len(split) > 2:
        _del_dict(type, iface, split[1], split[2], value)
    else:
        _del_col(type, iface, split[1], value)
    # Cleanup the tag itself (i.e. "netplan/column[/key]")
    subprocess.check_call([OVS_VSCTL_PATH, 'remove', type, iface, 'external-ids', setting])


def is_ovs_interface(iface, np_interface_dict):
    assert isinstance(np_interface_dict, dict)
    np_def = np_interface_dict.get(iface, None)
    return np_def and np_def.backend == 'OpenVSwitch'


def apply_ovs_cleanup(config_manager, ovs_old, ovs_current):  # pragma: nocover (covered in autopkgtest)
    """
    Query OpenVSwitch state through 'ovs-vsctl' and filter for netplan=true
    tagged ports/bonds and bridges. Delete interfaces which are not defined
    in the current configuration.
    Also filter for individual settings tagged netplan/<column>[/<key]=value
    in external-ids and clear them if they have been set by netplan.
    """
    if not _ovs_installed():
        raise OvsDbServerNotInstalled("Cannot apply OpenvSwitch cleanup: ovs-vsctl is 'not-found'")

    if not _ovs_active():
        raise OvsDbServerNotRunning('OpenvSwitch database is not running')

    config_manager.parse()
    ovs_ifaces = set()
    for i in config_manager.netdefs.keys():
        if (is_ovs_interface(i, config_manager.netdefs)):
            ovs_ifaces.add(i)

    # Tear down old OVS interfaces, not defined in the current config.
    # Use 'del-br' on the Interface table, to delete any netplan created VLAN fake bridges.
    # Use 'del-bond-iface' on the Interface table, to delete netplan created patch port interfaces
    # Step 1: Delete all interfaces, which are not part of the current OVS config
    for t in (('Port', 'del-port'), ('Bridge', 'del-br'), ('Interface', 'del-br')):
        out = subprocess.check_output([OVS_VSCTL_PATH, '--columns=name,external-ids',
                                       '-f', 'csv', '-d', 'bare', '--no-headings', 'list', t[0]],
                                      text=True)
        for line in out.splitlines():
            if 'netplan=true' in line:
                iface = line.split(',')[0]
                # Skip cleanup if this OVS interface is part of the current netplan OVS config
                if iface in ovs_ifaces:
                    continue
                if t[0] == 'Interface' and subprocess.run([OVS_VSCTL_PATH, 'br-exists', iface]).returncode > 0:
                    subprocess.check_call([OVS_VSCTL_PATH, '--if-exists', 'del-bond-iface', iface])
                else:
                    subprocess.check_call([OVS_VSCTL_PATH, '--if-exists', t[1], iface])

    # Step 2: Clean up the settings of the remaining interfaces
    for t in ('Port', 'Bridge', 'Interface', 'Open_vSwitch', 'Controller'):
        cols = 'name,external-ids'
        if t == 'Open_vSwitch':
            cols = 'external-ids'
        elif t == 'Controller':
            cols = '_uuid,external-ids'  # handle _uuid as if it would be the iface 'name'
        out = subprocess.check_output([OVS_VSCTL_PATH, '--columns=%s' % cols,
                                       '-f', 'csv', '-d', 'bare', '--no-headings', 'list', t],
                                      text=True)
        for line in out.splitlines():
            if 'netplan/' in line:
                iface = '.'
                extids = line
                if t != 'Open_vSwitch':
                    iface, extids = line.split(',', 1)
                # Check each line (interface) if it contains any netplan tagged settings, e.g.:
                # ovs0,"iface-id=myhostname netplan=true netplan/external-ids/iface-id=myhostname"
                # ovs1,"netplan=true netplan/global/set-fail-mode=standalone netplan/mcast_snooping_enable=false"
                for entry in extids.strip('"').split(' '):
                    if entry.startswith('netplan/') and '=' in entry:
                        setting, val = entry.split('=', 1)
                        clear_setting(t, iface, setting, val)
